"""
Module defining constants and enumerate types used in the package
"""
import functools
import mimetypes
import re
import warnings
from enum import Enum, EnumMeta
from typing import Any, Callable, Union

import numpy as np
import pyproj
import utm
from aenum import extend_enum

from ._version import __version__
from .exceptions import SHUserWarning


class PackageProps:
    """Class for obtaining package properties. Currently, it supports obtaining package version."""

    @staticmethod
    def get_version() -> str:
        """Returns package version

        :return: package version
        """
        return __version__


class ServiceUrl:
    """Most commonly used Sentinel Hub service URLs"""

    MAIN = "https://services.sentinel-hub.com"
    USWEST = "https://services-uswest2.sentinel-hub.com"
    CREODIAS = "https://creodias.sentinel-hub.com"
    MUNDI = "https://shservices.mundiwebservices.com"
    CODE_DE = "https://code-de.sentinel-hub.com"


class ServiceType(Enum):
    """Enum constant class for type of service

    Supported types are WMS, WCS, WFS, AWS, IMAGE
    """

    WMS = "wms"
    WCS = "wcs"
    WFS = "wfs"
    AWS = "aws"
    IMAGE = "image"
    FIS = "fis"
    PROCESSING_API = "processing"


class ResamplingTypeMeta(EnumMeta):
    """Metaclass for ResamplingType so that it is not case sensitive."""

    def __call__(cls, value: str, *args: Any, **kwargs: Any):  # type: ignore
        if isinstance(value, str):
            value = value.upper()

        return super().__call__(value, *args, **kwargs)


class ResamplingType(Enum, metaclass=ResamplingTypeMeta):
    """Enum constant class for type of resampling."""

    NEAREST = "NEAREST"
    BILINEAR = "BILINEAR"
    BICUBIC = "BICUBIC"


class MosaickingOrder(Enum):
    """Enum constant class for type of mosaicking order."""

    MOST_RECENT = "mostRecent"
    LEAST_RECENT = "leastRecent"
    LEAST_CC = "leastCC"


class CRSMeta(EnumMeta):
    """Metaclass used for building CRS Enum class"""

    _UNSUPPORTED_CRS = pyproj.CRS(4326)

    def __new__(mcs, cls, bases, classdict):  # type: ignore
        """This is executed at the beginning of runtime when CRS class is created"""
        for direction, direction_value in [("N", "6"), ("S", "7")]:
            for zone in range(1, 61):
                classdict[f"UTM_{zone}{direction}"] = f"32{direction_value}{zone:02}"

        return super().__new__(mcs, cls, bases, classdict)

    def __call__(cls, crs_value, *args, **kwargs):  # type: ignore
        """This is executed whenever CRS('something') is called"""
        # pylint: disable=signature-differs
        crs_value = cls._parse_crs(crs_value)

        if isinstance(crs_value, str) and not cls.has_value(crs_value) and crs_value.isdigit() and len(crs_value) >= 4:
            crs_name = f"EPSG_{crs_value}"
            extend_enum(cls, crs_name, crs_value)

        return super().__call__(crs_value, *args, **kwargs)

    @staticmethod
    def _parse_crs(value: Union[int, str, dict, pyproj.CRS]) -> str:
        """Method for parsing different inputs representing the same CRS enum. Examples:

        - 4326
        - 'EPSG:3857'
        - {'init': 32633}
        - geojson['crs']['properties']['name'] string (urn:ogc:def:crs:...)
        - pyproj.CRS(32743)
        """
        if isinstance(value, dict) and "init" in value:
            value = value["init"]
        if isinstance(value, pyproj.CRS):
            if value == CRSMeta._UNSUPPORTED_CRS:
                message = (
                    "sentinelhub-py supports only WGS 84 coordinate reference system with "
                    "coordinate order lng-lat. Given pyproj.CRS(4326) has coordinate order lat-lng. Be careful "
                    "to use the correct order of coordinates."
                )
                warnings.warn(message, category=SHUserWarning)

            epsg_code = value.to_epsg()
            if epsg_code is not None:
                return str(epsg_code)

            if value == CRS.WGS84.pyproj_crs():
                return "4326"

            error_message = f"Failed to determine an EPSG code of the given CRS:\n{repr(value)}"
            maybe_epsg = value.to_epsg(min_confidence=0)
            if maybe_epsg is not None:
                error_message = f"{error_message}\nIt might be EPSG {maybe_epsg} but pyproj is not confident enough."
            raise ValueError(error_message)

        if isinstance(value, (int, np.integer)):
            return str(value)
        if isinstance(value, str):
            if "urn:ogc:def:crs" in value.lower():
                crs_template = re.compile(r"urn:ogc:def:crs:.+::(?P<code>.+)", re.IGNORECASE)
                value = crs_template.match(value).group("code")  # type: ignore
            if value.upper() == "CRS84":
                return "4326"
            return value.lower().strip("epsg: ")
        return value  # type: ignore


class CRS(Enum, metaclass=CRSMeta):
    """Coordinate Reference System enumerate class

    Available CRS constants are WGS84, POP_WEB (i.e. Popular Web Mercator) and constants in form UTM_<zone><direction>,
    where zone is an integer from [1, 60] and direction is either N or S (i.e. northern or southern hemisphere)
    """

    WGS84 = "4326"
    POP_WEB = "3857"
    #: UTM enum members are defined in CRSMeta.__new__

    def __str__(self) -> str:
        """Method for casting CRS enum into string"""
        return self.ogc_string()

    def __repr__(self) -> str:
        """Method for retrieving CRS enum representation"""
        return f"CRS('{self.value}')"

    @classmethod
    def has_value(cls, value: str) -> bool:
        """Tests whether CRS contains a constant defined with string `value`.

        :param value: The string representation of the enum constant.
        :return: `True` if there exists a constant with string value `value`, `False` otherwise
        """
        return value in cls._value2member_map_

    @property
    def epsg(self) -> int:
        """EPSG code property

        :return: EPSG code of given CRS
        """
        return int(self.value)

    def ogc_string(self) -> str:
        """Returns a string of the form authority:id representing the CRS.

        :param self: An enum constant representing a coordinate reference system.
        :return: A string representation of the CRS.
        """
        return f"EPSG:{CRS(self).value}"

    @property
    def opengis_string(self) -> str:
        """Returns a URL to OGC webpage where the CRS is defined

        :return: A URL with CRS definition
        """
        return f"http://www.opengis.net/def/crs/EPSG/0/{self.epsg}"

    def is_utm(self) -> bool:
        """Checks if crs is one of the 64 possible UTM coordinate reference systems.

        :param self: An enum constant representing a coordinate reference system.
        :return: `True` if crs is UTM and `False` otherwise
        """
        return self.name.startswith("UTM")

    @functools.lru_cache(maxsize=128)
    def projection(self) -> pyproj.Proj:
        """Returns a projection in form of pyproj class.

        For better time performance this method will cache `128` most recent results. Cache can be released with
        `CRS.projection.cache_clear()`.

        :return: pyproj projection class
        """
        return pyproj.Proj(self._get_pyproj_projection_def(), preserve_units=True)

    @functools.lru_cache(maxsize=128)
    def pyproj_crs(self) -> pyproj.CRS:
        """Returns a pyproj CRS class.

        For better time performance this method will cache `128` most recent results. Cache can be released with
        `CRS.pyproj_crs.cache_clear()`.

        :return: pyproj CRS class
        """
        return pyproj.CRS(self._get_pyproj_projection_def())

    @functools.lru_cache(maxsize=512)
    def get_transform_function(self, other: "CRS", always_xy: bool = True) -> Callable[..., tuple]:
        """Returns a function for transforming geometrical objects from one CRS to another. The function will support
        transformations between any objects that pyproj supports.

        For better time performance this method will cache results. Cache can be released with
        `CRS.get_transform_function.cache_clear()`.

        :param self: Initial CRS
        :param other: Target CRS
        :param always_xy: Parameter that is passed to `pyproj.Transformer` object and defines axis order for
            transformation. The default value `True` is in most cases the correct one.
        :return: A projection function obtained from pyproj package
        """
        return pyproj.Transformer.from_proj(self.projection(), other.projection(), always_xy=always_xy).transform

    @staticmethod
    def get_utm_from_wgs84(lng: float, lat: float) -> "CRS":
        """Convert from WGS84 to UTM coordinate system

        :param lng: Longitude
        :param lat: Latitude
        :return: UTM coordinates
        """
        _, _, zone, _ = utm.from_latlon(lat, lng)
        direction = "N" if lat >= 0 else "S"
        return CRS[f"UTM_{zone}{direction}"]

    def _get_pyproj_projection_def(self) -> str:
        """Returns a pyproj crs definition

        For WGS 84 it ensures lng-lat order
        """
        return "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs" if self is CRS.WGS84 else self.ogc_string()


class CustomUrlParam(Enum):
    """Enum class to represent supported custom url parameters of OGC services

    Supported parameters are `SHOWLOGO`, `EVALSCRIPT`, `EVALSCRIPTURL`, `PREVIEW`, `QUALITY`, `UPSAMPLING`,
    `DOWNSAMPLING`, `GEOMETRY` and `WARNINGS`.

    See `documentation <https://www.sentinel-hub.com/develop/api/ogc/custom-parameters/>`__ for more information.
    """

    SHOWLOGO = "ShowLogo"
    EVALSCRIPT = "EvalScript"
    EVALSCRIPTURL = "EvalScriptUrl"
    PREVIEW = "Preview"
    QUALITY = "Quality"
    UPSAMPLING = "Upsampling"
    DOWNSAMPLING = "Downsampling"
    GEOMETRY = "Geometry"
    MINQA = "MinQA"

    @classmethod
    def has_value(cls, value: str) -> bool:
        """Tests whether CustomUrlParam contains a constant defined with a string `value`

        :param value: The string representation of the enum constant
        :return: `True` if there exists a constant with a string value `value`, `False` otherwise
        """
        return any(value.lower() == item.value.lower() for item in cls)

    @staticmethod
    def get_string(param: Enum) -> str:
        """Get custom url parameter name as string

        :param param: CustomUrlParam enum constant
        :return: String describing the file format
        """
        return param.value


class HistogramType(Enum):
    """Enum class for types of histogram supported by Sentinel Hub FIS service

    Supported histogram types are EQUALFREQUENCY, EQUIDISTANT and STREAMING
    """

    EQUALFREQUENCY = "equalfrequency"
    EQUIDISTANT = "equidistant"
    STREAMING = "streaming"


class MimeType(Enum):
    """Enum class to represent supported file formats

    Supported file formats are TIFF 8-bit, TIFF 16-bit, TIFF 32-bit float, PNG, JPEG, JPEG2000, JSON, CSV, ZIP, HDF5,
    XML, GML, RAW
    """

    TIFF = "tiff"
    PNG = "png"
    JPG = "jpg"
    JP2 = "jp2"
    JSON = "json"
    CSV = "csv"
    ZIP = "zip"
    HDF = "hdf"
    XML = "xml"
    GML = "gml"
    TXT = "txt"
    TAR = "tar"
    RAW = "raw"
    SAFE = "safe"
    PICKLE = "pkl"
    NPY = "npy"
    GPKG = "gpkg"
    GEOJSON = "geojson"
    GZIP = "gz"

    @property
    def extension(self) -> str:
        """Returns file extension of the MimeType object

        :returns: A file extension string
        """
        return self.value

    @staticmethod
    def from_string(mime_type_str: str) -> "MimeType":
        """Parses mime type from a file extension string

        :param mime_type_str: A file extension string
        :return: A mime type enum
        """
        guessed_extension = mimetypes.guess_extension(mime_type_str)
        if guessed_extension:
            mime_type_str = guessed_extension.strip(".")
        else:
            mime_type_str = mime_type_str.split("/")[-1]

        if MimeType.has_value(mime_type_str):
            return MimeType(mime_type_str)

        try:
            return {"tif": MimeType.TIFF, "jpeg": MimeType.JPG, "hdf5": MimeType.HDF, "h5": MimeType.HDF}[mime_type_str]
        except KeyError as exception:
            raise ValueError(f"Data format {mime_type_str} is not supported") from exception

    def is_image_format(self) -> bool:
        """Checks whether file format is an image format

        Example: ``MimeType.PNG.is_image_format()`` or ``MimeType.is_image_format(MimeType.PNG)``

        :param self: File format
        :return: `True` if file is in image format, `False` otherwise
        """
        return self in frozenset([MimeType.TIFF, MimeType.PNG, MimeType.JP2, MimeType.JPG])

    def is_api_format(self) -> bool:
        """Checks if mime type is supported by Sentinel Hub API

        :return: True if API supports this format and False otherwise
        """
        return self in frozenset([MimeType.JPG, MimeType.PNG, MimeType.TIFF, MimeType.JSON])

    @classmethod
    def has_value(cls, value: str) -> bool:
        """Tests whether MimeType contains a constant defined with string ``value``

        :param value: The string representation of the enum constant
        :return: `True` if there exists a constant with string value ``value``, `False` otherwise
        """
        return value in cls._value2member_map_

    def get_string(self) -> str:
        """Get file format as string

        :return: String describing the file format
        """
        if self is MimeType.JP2:
            return "image/jpeg2000"
        if self is MimeType.XML:
            return "text/xml"
        if self is MimeType.RAW:
            return self.value
        return mimetypes.types_map["." + self.value]

    def matches_extension(self, path: str) -> bool:
        """Checks if mime type enum is used as the last file extension in given file path.

        :param path: Path that might have an extension at the end.
        :return: A boolean value indicating if the file path ends with the expected extension.
        """
        return path.endswith(f".{self.extension}")

    def get_expected_max_value(self) -> Union[float, int]:
        """Returns max value of image `MimeType` format and raises an error if it is not an image format

        :return: A maximum value of specified image format
        :raises: ValueError
        """
        try:
            return {MimeType.TIFF: 65535, MimeType.PNG: 255, MimeType.JPG: 255, MimeType.JP2: 10000}[self]
        except KeyError as exception:
            raise ValueError(f"Type {self} is not supported by this method") from exception


class RequestType(Enum):
    """Enum constant class for GET/POST request type"""

    GET = "GET"
    POST = "POST"
    DELETE = "DELETE"
    PUT = "PUT"
    PATCH = "PATCH"


class SHConstants:
    """Initialisation of constants used by OGC request.

    Constants are LATEST
    """

    LATEST = "latest"
    HEADERS = {"User-Agent": f"sentinelhub-py/v{PackageProps.get_version()}"}



"""
Module that implements command line interface for the package
"""

from typing import Any, Callable, TypeVar

import click

from .config import SHConfig
from .download import DownloadClient, DownloadRequest

FC = TypeVar("FC", bound=Callable[..., Any])


@click.command()
def main_help() -> None:
    """
    Welcome to sentinelhub Python library command line help.

    \b
    There are multiple modules with command line functionality:\n
       - sentinelhub.aws \n
       - sentinelhub.config \n
       - sentinelhub.download \n

    To check more about certain module command use: \n
      sentinelhub.<module name> --help
    """


def _config_options(func: FC) -> FC:
    """A helper function which joins click.option functions of each parameter from config.json"""
    for param in SHConfig().get_params()[-1::-1]:
        func = click.option(f"--{param}", param, help=f'Set new values to configuration parameter "{param}"')(func)
    return func


@click.command()
@click.option("--show", is_flag=True, default=False, help="Show current configuration")
@click.option("--reset", is_flag=True, default=False, help="Reset configuration to initial state")
@_config_options
def config(show: bool, reset: bool, **params: Any) -> None:
    """Inspect and configure parameters in your local sentinelhub configuration file

    \b
    Example:
      sentinelhub.config --show
      sentinelhub.config --instance_id <new instance id>
      sentinelhub.config --max_download_attempts 5 --download_sleep_time 20 --download_timeout_seconds 120
    """
    sh_config = SHConfig()

    if reset:
        sh_config.reset()

    for param, value in params.items():
        if value is not None:
            try:
                value = int(value)
            except ValueError:
                if value.lower() == "true":
                    value = True
                elif value.lower() == "false":
                    value = False
            if getattr(sh_config, param) != value:
                setattr(sh_config, param, value)

    old_config = SHConfig()
    sh_config.save()

    for param in sh_config.get_params():
        if sh_config[param] != old_config[param]:
            value = sh_config[param]
            if isinstance(value, str):
                value = f"'{value}'"
            click.echo(f"The value of parameter '{param}' was updated to {value}")

    if show:
        click.echo(str(sh_config))
        click.echo(f"Configuration file location: {sh_config.get_config_location()}")


@click.command()
@click.argument("url")
@click.argument("filename", type=click.Path())
@click.option("-r", "--redownload", is_flag=True, default=False, help="Redownload existing files")
def download(url: str, filename: str, redownload: bool) -> None:
    """Download from custom created URL into custom created file path

    \b
    Example:
    sentinelhub.download https://roda.sentinel-hub.com/sentinel-s2-l1c/tiles/36/M/ZB/2022/3/17/0/metadata.xml \
./data/example.xml
    """
    data_folder, filename = filename.rsplit("/", 1)
    download_list = [
        DownloadRequest(url=url, data_folder=data_folder, filename=filename, save_response=True, return_data=False)
    ]
    DownloadClient(redownload=redownload).download(download_list)



"""
Download process for Sentinel Hub Statistical API
"""
import concurrent.futures
import copy
import logging
import os
from typing import Any, Dict, cast

from ..constants import MimeType
from ..decoding import decode_data as decode_data_function
from ..io_utils import read_data, write_data
from ..type_utils import JsonDict
from .request import DownloadRequest
from .sentinelhub_client import SentinelHubDownloadClient

LOGGER = logging.getLogger(__name__)


class SentinelHubStatisticalDownloadClient(SentinelHubDownloadClient):
    """A special download client for Sentinel Hub Statistical API

    Besides a normal download from Sentinel Hub services it implements an additional process of retrying and caching
    """

    _RETRIABLE_ERRORS = ["EXECUTION_ERROR", "TIMEOUT"]

    def __init__(self, *args: Any, n_interval_retries: int = 1, max_retry_threads: int = 5, **kwargs: Any):
        """
        :param n_interval_retries: Number of retries if a request fails just for a certain timestamp. (This parameter
            is experimental and might be changed in the future.)
        :param max_retry_threads: Number of threads used for retrying. (This parameter is experimental and might be
            changed in the future.)
        """
        super().__init__(*args, **kwargs)

        self.n_interval_retries = n_interval_retries
        self.max_retry_threads = max_retry_threads

    def _single_download(self, request: DownloadRequest, _: Any) -> Any:
        """Method for downloading a single request"""
        request.raise_if_invalid()
        if not (request.save_response or request.return_data):
            return None

        request_path, response_path = request.get_storage_paths()

        download_required = self.redownload or response_path is None or not os.path.exists(response_path)
        if download_required:
            response_content = self._execute_download(request)
            stats_response = decode_data_function(response_content, request.data_type)
        else:
            path = cast(str, response_path)  # fix in future, for now ensured by download_required
            LOGGER.debug("Reading locally stored data from %s instead of downloading", path)
            self._check_cached_request_is_matching(request, request_path)
            stats_response = read_data(path, data_format=request.data_type)

        failed_time_intervals: Dict[int, Any] = {}
        for index, stat_info in enumerate(stats_response["data"]):
            if self._has_retriable_error(stat_info):
                failed_time_intervals[index] = stat_info["interval"]

        n_succeeded_intervals = 0
        if failed_time_intervals:
            LOGGER.debug("Failed for %s intervals, retrying by downloading per interval", len(failed_time_intervals))
            retried_responses = self._download_per_interval(request, failed_time_intervals)
            n_succeeded_intervals = sum("error" not in stat_info for stat_info in retried_responses.values())

            stats_response["data"] = [
                retried_responses.get(index, stat_info) for index, stat_info in enumerate(stats_response["data"])
            ]

        if request_path and request.save_response and (self.redownload or not os.path.exists(request_path)):
            request_info = request.get_request_params(include_metadata=True)
            write_data(request_path, request_info, MimeType.JSON)
            LOGGER.debug("Saved request info to %s", request_path)

        if request.save_response and (download_required or n_succeeded_intervals > 0):
            path = cast(str, response_path)  # fix in future, for now ensured by raise_if_invalid
            write_data(path, stats_response, data_format=request.data_type)
            LOGGER.debug("Saved data to %s", response_path)

        if request.return_data:
            return stats_response
        return None

    def _download_per_interval(self, request: DownloadRequest, time_intervals: Dict[int, Any]) -> dict:
        """Download statistics per each time interval"""
        interval_requests = []
        for time_interval in time_intervals.values():
            interval_request = copy.deepcopy(request)
            if interval_request.post_values is None or "aggregation" not in interval_request.post_values:
                raise ValueError("Unable to configure request for retrying by interval.")

            interval_request.post_values["aggregation"]["timeRange"] = time_interval
            interval_requests.append(interval_request)

        with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_retry_threads) as executor:
            stat_info_responses = list(executor.map(self._execute_single_stat_download, interval_requests))

        return dict(zip(time_intervals, stat_info_responses))

    def _execute_single_stat_download(self, request: DownloadRequest) -> JsonDict:
        """Makes sure a download for a single time interval is retried"""
        for retry_count in range(self.n_interval_retries):
            response = self._execute_download(request)
            stat_response = decode_data_function(response, request.data_type)
            stat_info = stat_response["data"][0]

            if not self._has_retriable_error(stat_info) or retry_count == self.n_interval_retries - 1:
                return stat_info

        raise ValueError("No more retries available, download unsuccessful")

    def _has_retriable_error(self, stat_info: JsonDict) -> bool:
        """Checks if a dictionary of Stat API info for a single time interval has an error that can fixed by retrying
        a request
        """
        error_type = stat_info.get("error", {}).get("type")
        return error_type in self._RETRIABLE_ERRORS



"""
Utility functions to read/write image data from/to file
"""

import csv
import json
import logging
import os
import warnings
from typing import Any, Callable, Dict, Optional
from xml.etree import ElementTree

import numpy as np
import tifffile as tiff
from PIL import Image

from .constants import MimeType
from .decoding import decode_image_with_pillow, decode_jp2_image, decode_tar, get_data_format
from .exceptions import SHUserWarning
from .os_utils import create_parent_folder
from .type_utils import Json

LOGGER = logging.getLogger(__name__)

CSV_DELIMITER = ";"


def read_data(filename: str, data_format: Optional[MimeType] = None) -> Any:
    """Read image data from file

    This function reads input data from file. The format of the file
    can be specified in ``data_format``. If not specified, the format is
    guessed from the extension of the filename.

    :param filename: filename to read data from
    :param data_format: format of filename. Default is `None`
    :return: data read from filename
    :raises: exception if filename does not exist
    """
    if not os.path.exists(filename):
        raise FileNotFoundError(f"Filename {filename} does not exist")

    if not isinstance(data_format, MimeType):
        data_format = get_data_format(filename)

    reader = _get_reader(data_format)

    try:
        return reader(filename)
    except BaseException as exception:
        # In case a procedure would read a lot of files and one would be corrupt this helps us figure out which one
        LOGGER.debug("Failed to read from file: %s", filename)
        raise exception


def _get_reader(data_format: MimeType) -> Callable[[str], Any]:
    """Provides a function for reading data in a given data format"""
    if data_format is MimeType.TIFF:
        return read_tiff_image
    if data_format is MimeType.JP2:
        return read_jp2_image
    if data_format.is_image_format():
        return read_image
    try:
        available_readers: Dict[MimeType, Callable[[str], Any]] = {
            MimeType.TAR: read_tar,
            MimeType.TXT: read_text,
            MimeType.RAW: _read_binary,
            MimeType.CSV: read_csv,
            MimeType.JSON: read_json,
            MimeType.XML: read_xml,
            MimeType.GML: read_xml,
            MimeType.SAFE: read_xml,
        }
        return available_readers[data_format]
    except KeyError as exception:
        raise ValueError(f"Reading data format {data_format} is not supported") from exception


def read_tar(filename: str) -> Dict[str, object]:
    """Read a tar from file"""
    with open(filename, "rb") as file:
        return decode_tar(file)  # type: ignore


def read_tiff_image(filename: str) -> Any:
    """Read data from TIFF file

    :param filename: name of TIFF file to be read
    :return: data stored in TIFF file
    """
    return tiff.imread(filename)


def read_jp2_image(filename: str) -> np.ndarray:
    """Read data from JPEG2000 file

    :param filename: name of JPEG2000 file to be read
    :return: data stored in JPEG2000 file
    """
    with open(filename, "rb") as file:
        return decode_jp2_image(file)


def read_image(filename: str) -> np.ndarray:
    """Read data from PNG or JPG file

    :param filename: name of PNG or JPG file to be read
    :return: data stored in JPG file
    """
    return decode_image_with_pillow(filename)


def read_text(filename: str) -> str:
    """Read data from text file

    :param filename: name of text file to be read
    :return: data stored in text file
    """
    with open(filename, "r") as file:
        return file.read()


def _read_binary(filename: str) -> bytes:
    """Reads data in bytes"""
    with open(filename, "rb") as file:
        return file.read()


def read_csv(filename: str, delimiter: str = CSV_DELIMITER) -> list:
    """Read data from CSV file

    :param filename: name of CSV file to be read
    :param delimiter: type of CSV delimiter. Default is ``;``
    :return: data stored in CSV file as list
    """
    with open(filename, "r") as file:
        return list(csv.reader(file, delimiter=delimiter))


def read_json(filename: str) -> Any:
    """Read data from JSON file

    :param filename: name of JSON file to be read
    :return: data stored in JSON file
    """
    with open(filename, "rb") as file:
        return json.load(file)


def read_xml(filename: str) -> ElementTree.ElementTree:
    """Read data from XML or GML file

    :param filename: name of XML or GML file to be read
    :return: data stored in XML file
    """
    return ElementTree.parse(filename)


def read_numpy(filename: str) -> np.ndarray:
    """Read data from numpy file

    :param filename: name of numpy file to be read
    :return: data stored in file as numpy array
    """
    return np.load(filename)


def write_data(
    filename: str, data: Any, data_format: Optional[MimeType] = None, compress: bool = False, add: bool = False
) -> None:
    """Write image data to file

    Function to write image data to specified file. If file format is not provided
    explicitly, it is guessed from the filename extension. If format is TIFF, geo
    information and compression can be optionally added.

    :param filename: name of file to write data to
    :param data: image data to write to file
    :param data_format: format of output file. Default is `None`
    :param compress: whether to compress data or not. Default is `False`
    :param add: whether to append to existing text file or not. Default is `False`
    :raises: exception if numpy format is not supported or file cannot be written
    """
    create_parent_folder(filename)

    if not isinstance(data_format, MimeType):
        data_format = get_data_format(filename)

    if data_format is MimeType.TIFF:
        return write_tiff_image(filename, data, compress)
    if data_format.is_image_format():
        return write_image(filename, data)
    if data_format is MimeType.TXT:
        return write_text(filename, data, add=add)

    try:
        available_writers: Dict[MimeType, Callable[[str, Any], None]] = {
            MimeType.RAW: write_bytes,
            MimeType.CSV: write_csv,
            MimeType.JSON: write_json,
            MimeType.XML: write_xml,
            MimeType.GML: write_xml,
        }
        return available_writers[data_format](filename, data)
    except KeyError as exception:
        raise ValueError(f"Writing data format {data_format} is not supported") from exception


def write_tiff_image(filename: str, image: np.ndarray, compress: bool = False) -> None:
    """Write image data to TIFF file

    :param filename: name of file to write data to
    :param image: image data to write to file
    :param compress: whether to compress data. If `True`, lzma compression is used. Default is `False`
    """
    if compress:
        return tiff.imwrite(filename, image, compression="lzma")  # lossless compression, works very well on masks
    return tiff.imwrite(filename, image)


def write_jp2_image(filename: str, image: np.ndarray) -> None:
    """Write image data to JPEG2000 file

    :param filename: name of JPEG2000 file to write data to
    :param image: image data to write to file
    """
    # Other options:
    # return glymur.Jp2k(filename, data=image)
    # cv2.imwrite(filename, image)
    return write_image(filename, image)


def write_image(filename: str, image: np.ndarray) -> None:
    """Write image data to PNG, JPG file

    :param filename: name of PNG or JPG file to write data to
    :param image: image data to write to file
    """
    data_format = get_data_format(filename)
    if data_format is MimeType.JPG:
        warnings.warn("JPEG is a lossy format therefore saved data will be modified.", category=SHUserWarning)
    return Image.fromarray(image).save(filename)


def write_text(filename: str, data: np.ndarray, add: bool = False) -> None:
    """Write image data to text file

    :param filename: name of text file to write data to
    :param data: image data to write to text file
    :param add: whether to append to existing file or not. Default is `False`
    """
    write_type = "a" if add else "w"
    with open(filename, write_type) as file:
        print(data, end="", file=file)


def write_csv(filename: str, data: np.ndarray, delimiter: str = CSV_DELIMITER) -> None:
    """Write image data to CSV file

    :param filename: name of CSV file to write data to
    :param data: image data to write to CSV file
    :param delimiter: delimiter used in CSV file. Default is ``;``
    """
    with open(filename, "w") as file:
        csv_writer = csv.writer(file, delimiter=delimiter)
        for line in data:
            csv_writer.writerow(line)


def write_json(filename: str, data: Json) -> None:
    """Write data to JSON file

    :param filename: name of JSON file to write data to
    :param data: data to write to JSON file
    """
    with open(filename, "w") as file:
        json.dump(data, file, indent=4, sort_keys=True)


def write_xml(filename: str, element_tree: ElementTree.ElementTree) -> None:
    """Write data to XML or GML file

    :param filename: name of XML or GML file to write data to
    :param element_tree: data as ElementTree object
    """
    return element_tree.write(filename)
    # this will write declaration tag in first line:
    # return element_tree.write(filename, encoding='utf-8', xml_declaration=True)


def write_numpy(filename: str, data: np.ndarray) -> None:
    """Write data as numpy file

    :param filename: name of numpy file to write data to
    :param data: data to write to numpy file
    """
    return np.save(filename, data)


def write_bytes(filename: str, data: bytes) -> None:
    """Write binary data into a file

    :param filename: name of file to write the data to
    :param data: binary data to write
    """
    with open(filename, "wb") as file:
        file.write(data)



"""
Module implementing rate limiting logic for Sentinel Hub service
"""
import time
from enum import Enum
from typing import Union

from ..type_utils import JsonDict


class PolicyType(Enum):
    """Enum defining different types of policies"""

    PROCESSING_UNITS = "PROCESSING_UNITS"
    REQUESTS = "REQUESTS"


class SentinelHubRateLimit:
    """Class implementing rate limiting logic of Sentinel Hub service

    It has 2 public methods:

    - register_next - tells if next download can start or if not, what is the wait before it can be asked again
    - update - updates expectations according to headers obtained from download

    The rate limiting object is collecting information about the status of rate limiting policy buckets from
    Sentinel Hub service. According to this information and a feedback from download requests it adapts expectations
    about when the next download attempt will be possible.
    """

    REQUEST_RETRY_HEADER = "Retry-After"
    REQUEST_COUNT_HEADER = "X-RateLimit-Remaining"
    UNITS_RETRY_HEADER = "X-ProcessingUnits-Retry-After"
    UNITS_COUNT_HEADER = "X-ProcessingUnits-Remaining"
    VIOLATION_HEADER = "X-RateLimit-ViolatedPolicy"

    def __init__(self, num_processes: int = 1, minimum_wait_time: float = 0.05, maximum_wait_time: float = 60.0):
        """
        :param num_processes: Number of parallel download processes running.
        :param minimum_wait_time: Minimum wait time between two consecutive download requests in seconds.
        :param maximum_wait_time: Maximum wait time between two consecutive download requests in seconds.
        """
        self.wait_time = min(num_processes * minimum_wait_time, maximum_wait_time)
        self.next_download_time = time.monotonic()

    def register_next(self) -> float:
        """Determines if next download request can start or not by returning the waiting time in seconds."""
        current_time = time.monotonic()
        wait_time = max(self.next_download_time - current_time, 0)

        if wait_time == 0:
            self.next_download_time = max(current_time + self.wait_time, self.next_download_time)

        return wait_time

    def update(self, headers: dict) -> None:
        """Update the next possible download time if the service has responded with the rate limit"""
        retry_after: float
        retry_after = max(int(headers.get(self.REQUEST_RETRY_HEADER, 0)), int(headers.get(self.UNITS_RETRY_HEADER, 0)))
        retry_after = retry_after / 1000

        if retry_after:
            self.next_download_time = max(time.monotonic() + retry_after, self.next_download_time)


class PolicyBucket:
    """A class representing Sentinel Hub policy bucket"""

    def __init__(self, policy_type: Union[str, PolicyType], policy_payload: JsonDict):
        """
        :param policy_type: A type of policy
        :param policy_payload: A dictionary of policy parameters
        """

        self.policy_type = PolicyType(policy_type)

        self.capacity = float(policy_payload["capacity"])
        self.refill_period = policy_payload["samplingPeriod"]

        # The following is the same as if we would interpret samplingPeriod string
        self.refill_per_second = 10**9 / policy_payload["nanosBetweenRefills"]

        self._content = self.capacity

    def __repr__(self) -> str:
        """Representation of the bucket content"""
        return (
            f"{self.__class__.__name__}(policy_type={self.policy_type}, content={self.content}/{self.capacity}, "
            f"refill_period={self.refill_period}, refill_per_second={self.refill_per_second})"
        )

    @property
    def content(self) -> float:
        """Variable `content` can be accessed as a property"""
        return self._content

    @content.setter
    def content(self, value: float) -> None:
        """Variable `content` can be modified by external classes"""
        self._content = value

    def count_cost_per_second(self, elapsed_time: float, new_content: float) -> float:
        """Calculates the cost per second for the bucket given the elapsed time and the new content.

        In the calculation it assumes that during the elapsed time bucket was being filled all the time - i.e. it
        assumes the bucket has never been full for a non-zero amount of time in the elapsed time period.
        """
        content_difference = self.content - new_content
        if not self.is_fixed():
            content_difference += elapsed_time * self.refill_per_second

        return content_difference / elapsed_time

    def get_wait_time(
        self,
        elapsed_time: float,
        process_num: int,
        cost_per_request: float,
        requests_completed: int,
        buffer_cost: float = 0.5,
    ) -> float:
        """Expected time a user would have to wait for this bucket"""
        overall_completed_cost = requests_completed * cost_per_request * process_num
        expected_content = max(self.content + elapsed_time * self.refill_per_second - overall_completed_cost, 0)

        if self.is_fixed():
            if expected_content < cost_per_request:
                return -1
            return 0

        return max(cost_per_request - expected_content + buffer_cost, 0) / self.refill_per_second

    def is_request_bucket(self) -> bool:
        """Checks if bucket counts requests"""
        return self.policy_type is PolicyType.REQUESTS

    def is_fixed(self) -> bool:
        """Checks if bucket has a fixed number of requests"""
        return self.refill_period == "PT0S"



"""
Module implementing a download client that is adjusted to download from AWS
"""
import logging
import warnings
from typing import Any

from ..download.request import DownloadRequest

try:
    import boto3
    from botocore.exceptions import NoCredentialsError
except ImportError as import_exception:
    raise ImportError(
        "To use AWS functionalities of this package you have to install sentinelhub[AWS] package extension"
    ) from import_exception

from ..download.client import DownloadClient
from ..download.handlers import fail_missing_file
from ..exceptions import AwsDownloadFailedException

LOGGER = logging.getLogger(__name__)


class AwsDownloadClient(DownloadClient):
    """An AWS download client class"""

    GLOBAL_S3_CLIENT = None

    @fail_missing_file
    def _execute_download(self, request: DownloadRequest) -> Any:
        """Executes a download procedure"""
        if not self.is_s3_request(request):
            return super()._execute_download(request)

        s3_client = self._get_s3_client()

        response_content = self._do_download(request, s3_client)

        LOGGER.debug("Successful download from %s", request.url)
        return response_content

    def _get_s3_client(self) -> Any:
        """Provides a s3 client object"""
        warnings.filterwarnings("ignore", category=ResourceWarning, message="unclosed.*<ssl.SSLSocket.*>")
        try:
            s3_client = boto3.Session().client(
                "s3",
                aws_access_key_id=self.config.aws_access_key_id or None,
                aws_secret_access_key=self.config.aws_secret_access_key or None,
                aws_session_token=self.config.aws_session_token or None,
            )
            AwsDownloadClient.GLOBAL_S3_CLIENT = s3_client
        except KeyError as exception:  # Sometimes creation of client fails, and we use the global client if it exists
            if AwsDownloadClient.GLOBAL_S3_CLIENT is None:
                raise ValueError("Failed to create a client for download from AWS") from exception
            s3_client = AwsDownloadClient.GLOBAL_S3_CLIENT

        return s3_client

    @staticmethod
    def _do_download(request: DownloadRequest, s3_client: Any) -> Any:
        """Does the download from s3"""
        if request.url is None:
            raise ValueError(f"Faulty request {request}, no URL specified.")
        _, _, bucket_name, url_key = request.url.split("/", 3)

        try:
            response = s3_client.get_object(Bucket=bucket_name, Key=url_key, RequestPayer="requester")
            return response["Body"].read()
        except NoCredentialsError as exception:
            raise ValueError(
                "The requested data is in Requester Pays AWS bucket. In order to download the data please set "
                "your access key either in AWS credentials file or in sentinelhub config.json file using "
                "command line:\n"
                "$ sentinelhub.config --aws_access_key_id <your AWS key> --aws_secret_access_key "
                "<your AWS secret key>"
            ) from exception
        except s3_client.exceptions.NoSuchKey as exception:
            raise AwsDownloadFailedException(f"File in location {request.url} is missing") from exception
        except s3_client.exceptions.NoSuchBucket as exception:
            raise ValueError(f"Aws bucket {bucket_name} does not exist") from exception

    @staticmethod
    def is_s3_request(request: DownloadRequest) -> bool:
        """Checks if data has to be downloaded from AWS s3 bucket

        :return: `True` if url describes location at AWS s3 bucket and `False` otherwise
        """
        return request.url is not None and request.url.startswith("s3://")



import io
import os
from typing import List

from setuptools import find_packages, setup


def parse_requirements(filename: str) -> List[str]:
    required_packages = []
    with open(os.path.join(os.path.dirname(__file__), filename)) as req_file:
        for line in req_file:
            required_packages.append(line.strip())
    return required_packages


def get_version() -> str:
    path = os.path.join(os.path.dirname(__file__), "sentinelhub", "_version.py")
    for line in open(path):
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            return version.strip('"').strip("'")

    raise ValueError(f"Version not found in {path}")


def get_long_description() -> str:
    return io.open("README.md", encoding="utf-8").read()


def update_package_config() -> None:
    """Every time sentinelhub package is installed entire config.json is overwritten. However, this function
    will check if sentinelhub is already installed and try to copy those parameters from old config.json that are by
    default set to an empty value (i.e. instance_id, aws_access_key_id and aws_secret_access_key) into new config.json
    file.
    """
    try:
        import importlib
        import json
        import sys

        spec = importlib.machinery.PathFinder().find_spec("sentinelhub", sys.path[1:])
        if spec is None or spec.submodule_search_locations is None:
            return

        path = spec.submodule_search_locations[0]
        old_config_filename = os.path.join(path, "config.json")

        with open(old_config_filename, "r") as file:
            old_config = json.load(file)

        from sentinelhub.config import SHConfig

        config = SHConfig()
        for attr, value in old_config.items():
            if hasattr(config, attr) and not getattr(config, attr):
                setattr(config, attr, value)

        config.save()

    except BaseException:
        pass


def try_create_config_file() -> None:
    """After the package is installed it will try to trigger saving a config.json file"""
    try:
        from sentinelhub.config import SHConfig

        SHConfig()
    except BaseException:
        pass


update_package_config()

setup(
    name="sentinelhub",
    python_requires=">=3.7",
    version=get_version(),
    description="Sentinel Hub Utilities",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/sentinel-hub/sentinelhub-py",
    author="Sinergise ltd.",
    author_email="info@sentinel-hub.com",
    license="MIT",
    packages=find_packages(),
    package_data={
        "sentinelhub": [
            "sentinelhub/config.json",
            "sentinelhub/.utmzones.geojson",
            "sentinelhub/py.typed",
        ]
    },
    include_package_data=True,
    install_requires=parse_requirements("requirements.txt"),
    extras_require={
        "AWS": parse_requirements("requirements-aws.txt"),
        "DEV": parse_requirements("requirements-dev.txt"),
        "DOCS": parse_requirements("requirements-docs.txt"),
    },
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "sentinelhub=sentinelhub.commands:main_help",
            "sentinelhub.aws=sentinelhub.aws.commands:aws",
            "sentinelhub.config=sentinelhub.commands:config",
            "sentinelhub.download=sentinelhub.commands:download",
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
    ],
)
try_create_config_file()



"""
Module implementing Sentinel Hub session object
"""
import base64
import json
import logging
import sys
import time
import warnings
from threading import Event, Thread
from typing import Any, Dict, Optional

import requests
from oauthlib.oauth2 import BackendApplicationClient
from requests import Response
from requests.exceptions import JSONDecodeError
from requests_oauthlib import OAuth2Session

from ..config import SHConfig
from ..download.handlers import fail_user_errors, retry_temporary_errors
from ..download.request import DownloadRequest
from ..exceptions import SHUserWarning
from ..type_utils import JsonDict

if sys.version_info < (3, 8):
    from shared_memory import SharedMemory
else:
    from multiprocessing.shared_memory import SharedMemory


LOGGER = logging.getLogger(__name__)


class SentinelHubSession:
    """Sentinel Hub authentication class

    The class will do OAuth2 authentication with Sentinel Hub service and store the token. It is able to refresh the
    token before it expires and decode user information from the token.

    For more info about Sentinel Hub authentication check
    `service documentation <https://docs.sentinel-hub.com/api/latest/api/overview/authentication/>`__.
    """

    DEFAULT_SECONDS_BEFORE_EXPIRY = 120

    def __init__(
        self,
        config: Optional[SHConfig] = None,
        refresh_before_expiry: Optional[float] = DEFAULT_SECONDS_BEFORE_EXPIRY,
        *,
        _token: Optional[JsonDict] = None,
    ):
        """
        :param config: A config object containing Sentinel Hub OAuth credentials and the base URL of the service.
        :param refresh_before_expiry: A number of seconds before authentication token expiry at which time a refreshing
            mechanism is activated. When this is activated it means that whenever a valid token will be again
            required the `SentinelHubSession` will re-authenticate to Sentinel Hub service and obtain a new token.
            By default, the parameter is set to `60` seconds. If this parameter is set to `None` it will deactivate
            token refreshing and `SentinelHubSession` might provide a token that is already expired. This can be used
            to avoid re-authenticating too many times.
        """
        self.config = config or SHConfig()
        self.refresh_before_expiry = refresh_before_expiry

        token_fetching_required = _token is None or self.refresh_before_expiry is not None
        if token_fetching_required and not (self.config.sh_client_id and self.config.sh_client_secret):
            raise ValueError(
                "Configuration parameters 'sh_client_id' and 'sh_client_secret' have to be set in order "
                "to authenticate with Sentinel Hub service. Check "
                "https://sentinelhub-py.readthedocs.io/en/latest/configure.html for more info."
            )

        self._token = self._collect_new_token() if _token is None else _token

    @classmethod
    def from_token(cls, token: JsonDict) -> "SentinelHubSession":
        """Create a session object from the given token. The created session is configured not to refresh its token.

        :param token: A dictionary containing token object.
        """
        for key in ["access_token", "expires_at"]:
            if key not in token:
                raise ValueError(f"Given token should be a dictionary containing a key '{key}'")

        return cls(_token=token, refresh_before_expiry=None)

    @property
    def token(self) -> JsonDict:
        """Always up-to-date session's token

        :return: A token in a form of dictionary of parameters
        """
        remaining_token_time = self._token["expires_at"] - time.time()
        if self.refresh_before_expiry is None:
            if remaining_token_time <= 0:
                warnings.warn("The Sentinel Hub session token seems to be expired.", category=SHUserWarning)
            return self._token

        if remaining_token_time <= self.refresh_before_expiry:
            self._token = self._collect_new_token()

        return self._token

    def info(self) -> JsonDict:
        """Decode token to get token info"""

        token = self.token["access_token"].split(".")[1]
        padded = token + "=" * (len(token) % 4)
        decoded_string = base64.b64decode(padded).decode()
        return json.loads(decoded_string)

    @property
    def session_headers(self) -> Dict[str, str]:
        """Provides session authorization headers

        :return: A dictionary with authorization headers.
        """
        return {"Authorization": f'Bearer {self.token["access_token"]}'}

    def _collect_new_token(self) -> JsonDict:
        """Creates a download request and fetches a token from the service.

        Note that the `DownloadRequest` object is created only because retry decorators of `_fetch_token` method
        require it.
        """
        request = DownloadRequest(url=self.config.get_sh_oauth_url())
        return self._fetch_token(request)

    @retry_temporary_errors
    @fail_user_errors
    def _fetch_token(self, request: DownloadRequest) -> JsonDict:
        """Collects a new token from Sentinel Hub service"""
        oauth_client = BackendApplicationClient(client_id=self.config.sh_client_id)

        LOGGER.debug("Creating a new authentication session with Sentinel Hub service")
        with OAuth2Session(client=oauth_client) as oauth_session:
            oauth_session.register_compliance_hook("access_token_response", self._compliance_hook)

            return oauth_session.fetch_token(
                token_url=request.url, client_id=self.config.sh_client_id, client_secret=self.config.sh_client_secret
            )

    @staticmethod
    def _compliance_hook(response: Response) -> Response:
        """Checks if a response from Sentinel Hub Authentication service has an error status code but no error message.

        By default, `requests_oauthlib` ignores status of a response and only looks at an error message in a
        response body. However, Sentinel Hub service can return a response with an error status code and
        without an error message. In such cases `requests_oauthlib` would raise a completely wrong error message. This
        hook makes sure that a correct error message is raised.

        It is important that in case of 5xx errors an error is always raised so that authentication can be retried.
        But in case of 4xx errors where response contains an error message this method intentionally doesn't raise
        an error so that `oauthlib` can later raise a more descriptive error.
        """
        if response.status_code >= requests.status_codes.codes.INTERNAL_SERVER_ERROR:
            response.raise_for_status()

        try:
            token_dict = response.json()
            if "error" in token_dict:
                return response
        except JSONDecodeError:
            pass

        response.raise_for_status()
        return response


_DEFAULT_SESSION_MEMORY_NAME = "sh-session-token"
_NULL_MEMORY_VALUE = b"\x00"


class SessionSharingThread(Thread):
    """A thread for sharing a token from `SentinelHubSession` object in a shared memory object that can be accessed by
    other Python processes during multiprocessing parallelization.

    How to use it:

    .. code-block:: python

        thread = SessionSharingThread(session)
        thread.start()

        # Run a parallelization process here
        # Use collect_shared_session() to retrieve the session with other processes

        thread.join()
    """

    _EXTRA_MEMORY_BYTES = 100

    def __init__(self, session: SentinelHubSession, memory_name: str = _DEFAULT_SESSION_MEMORY_NAME, **kwargs: Any):
        """
        :param session: A Sentinel Hub session to be used for sharing its authentication token.
        :param memory_name: A unique name for the requested shared memory block.
        :param kwargs: Keyword arguments to be propagated to `threading.Thread` parent class.
        """
        super().__init__(**kwargs)

        self.session = session
        self.memory_name = memory_name

        if self.session.refresh_before_expiry is None:
            raise ValueError(f"Given instance of {self.session.__class__.__name__} must be self-refreshing")
        self._refresh_time = self.session.refresh_before_expiry

        self._stop_event = Event()
        self._is_memory_shared_event = Event()

    def start(self) -> None:
        """Start running the thread.

        After starting the thread it also waits for the token to be shared. This way no other process would try to
        access the memory before it even exists."""
        super().start()
        self._is_memory_shared_event.wait()

    def run(self) -> None:
        """A running thread is running an infinite loop of sharing a token and waiting for token to expire. The loop
        ends only when the thread is stopped."""
        self._stop_event.clear()

        while not self._stop_event.is_set():
            token = self.session.token
            self._share_token(token)

            sleep_until_refresh_time = token["expires_at"] - time.time() - self._refresh_time
            if sleep_until_refresh_time > 0:
                self._stop_event.wait(timeout=sleep_until_refresh_time)

    def _share_token(self, token: JsonDict) -> None:
        """A token is encoded into bytes and written into a shared memory block."""
        encoded_token = json.dumps(token).encode()
        memory = self._get_shared_memory(encoded_token)

        try:
            memory.buf[:] = encoded_token + _NULL_MEMORY_VALUE * (memory.size - len(encoded_token))
        finally:
            memory.close()

    def _get_shared_memory(self, encoded_token: bytes) -> SharedMemory:
        """Provides a shared memory object.

        The method also handles a case where a shared memory with the same name would be left unclosed from before.
        Because the memory can be persistent and requires low-level knowledge of `multiprocessing.shared_memory` to
        close it manually this method will close it automatically and inform users about the problem.
        """
        if self._is_memory_shared_event.is_set():
            return SharedMemory(name=self.memory_name)

        try:
            memory = self._create_shared_memory(encoded_token)
        except FileExistsError:
            warnings.warn(
                f"A shared memory with a name '{self.memory_name}' already exists. It will be removed and allocated "
                f"anew. Please make sure that every {self.__class__.__name__} instance is joined at the end. If you "
                "are using multiple threads then specify different 'memory_name' parameter for each of them.",
                category=SHUserWarning,
            )

            memory = SharedMemory(name=self.memory_name)
            memory.unlink()
            memory.close()

            memory = self._create_shared_memory(encoded_token)

        self._is_memory_shared_event.set()
        return memory

    def _create_shared_memory(self, encoded_token: bytes) -> SharedMemory:
        """Create a new shared memory space.

        Note that the `SharedMemory` object allocates extra `self._EXTRA_MEMORY_BYTES` bytes of memory because the
        length of encoded token can vary a bit.
        """
        return SharedMemory(
            create=True,
            size=len(encoded_token) + self._EXTRA_MEMORY_BYTES,
            name=self.memory_name,
        )

    def join(self, timeout: Optional[float] = None) -> None:
        """The method stops the thread that would otherwise run indefinitely and joins it with the main thread.

        :param timeout: Parameter that is propagated to `threading.Thread.join` method.
        """
        self._stop_event.set()
        super().join(timeout=timeout)

        if self._is_memory_shared_event.is_set():
            try:
                memory = SharedMemory(name=self.memory_name)
                memory.unlink()
                memory.close()
            except FileNotFoundError:
                pass

            self._is_memory_shared_event.clear()


class SessionSharing:
    """An object that in the background runs a `SessionSharingThread` which shares a Sentinel Hub authentication
    token in a shared memory object that can be accessed by other Python processes during multiprocessing
    parallelization. The object also makes sure that the thread is always closed at the end.

    How to use it:

    .. code-block:: python

        with SessionSharing(session):
            # Run a parallelization process here
    """

    def __init__(self, session: SentinelHubSession, **kwargs: Any):
        """
        :param args: A Sentinel Hub session to be used for sharing its authentication token.
        :param kwargs: Keyword arguments to be propagated to `SessionSharingThread`.
        """
        self.thread = SessionSharingThread(session, **kwargs)

    def __enter__(self) -> None:
        """Starts running the session-sharing thread."""
        self.thread.start()

    def __exit__(self, *_: Any, **__: Any) -> None:
        """Closes the running session-sharing thread."""
        self.thread.join()


def collect_shared_session(memory_name: str = _DEFAULT_SESSION_MEMORY_NAME) -> SentinelHubSession:
    """This utility function is meant to be used in combination with `SessionSharingThread`. It retrieves an
    authentication token from the shared memory and returns it in an `SentinelHubSession` object.

    :param memory_name: A unique name of the requested shared memory block from where to read the session. It should
        match the one used in `SessionSharingThread`.
    :return: An instance of `SentinelHubSession` that contains the shared token but is not self-refreshing.
    """
    try:
        memory = SharedMemory(name=memory_name)
    except FileNotFoundError as exception:
        raise FileNotFoundError(
            f"Couldn't obtain a shared session because a shared memory '{memory_name}' doesn't exist. Make sure that"
            " you are running session sharing when calling this function"
        ) from exception

    try:
        encoded_token = memory.buf.tobytes().rstrip(_NULL_MEMORY_VALUE)
    finally:
        memory.close()

    token: JsonDict = json.loads(encoded_token)
    return SentinelHubSession.from_token(token)



"""
Module defining data collections
"""
import warnings
from dataclasses import dataclass, field, fields
from enum import Enum, EnumMeta
from typing import Any, Dict, List, Optional, Tuple

from aenum import extend_enum

from .config import SHConfig
from .constants import ServiceUrl
from .data_collections_bands import Band, Bands, MetaBands
from .exceptions import SHDeprecationWarning


class _CollectionType:
    """Types of Sentinel Hub data collections"""

    SENTINEL1 = "Sentinel-1"
    SENTINEL2 = "Sentinel-2"
    SENTINEL3 = "Sentinel-3"
    SENTINEL5P = "Sentinel-5P"
    LANDSAT_MSS = "Landsat 1-5 MSS"
    LANDSAT_TM = "Landsat 4-5 TM"
    LANDSAT5 = "Landsat 5"
    LANDSAT_ETM = "Landsat 7 ETM+"
    LANDSAT_OT = "Landsat 8 OLI and TIRS"
    MODIS = "MODIS"
    ENVISAT_MERIS = "Envisat Meris"
    DEM = "DEM"
    BYOC = "BYOC"
    BATCH = "BATCH"


class _SensorType:
    """Satellite sensors"""

    # pylint: disable=invalid-name
    MSI = "MSI"
    OLI_TIRS = "OLI-TIRS"
    TM = "TM"
    ETM = "ETM+"
    MSS = "MSS"
    C_SAR = "C-SAR"
    OLCI = "OLCI"
    SLSTR = "SLSTR"
    TROPOMI = "TROPOMI"


class _ProcessingLevel:
    """Processing levels"""

    # pylint: disable=invalid-name
    L1 = "L1"
    L2 = "L2"
    L1B = "L1B"
    L1C = "L1C"
    L2A = "L2A"
    L3B = "L3B"
    GRD = "GRD"


class _SwathMode:
    """Swath modes for SAR sensors"""

    # pylint: disable=invalid-name
    IW = "IW"
    EW = "EW"
    SM = "SM"
    WV = "WV"


class _Polarization:
    """SAR polarizations"""

    # pylint: disable=invalid-name
    DV = "DV"
    DH = "DH"
    SV = "SV"
    SH = "SH"
    HH = "HH"
    HV = "HV"
    VV = "VV"
    VH = "VH"


class _Resolution:
    """Product resolution (specific to Sentinel-1 collections)"""

    MEDIUM = "MEDIUM"
    HIGH = "HIGH"


class OrbitDirection:
    """Orbit directions"""

    ASCENDING = "ASCENDING"
    DESCENDING = "DESCENDING"
    BOTH = "BOTH"


def _shallow_asdict(dataclass_instance: Any) -> Dict[str, Any]:
    """Returns a dictionary of fields and values, but is not recursive and does not deepcopy like `asdict`"""
    # This definition needs to be above the class definitions in the file
    return {field.name: getattr(dataclass_instance, field.name) for field in fields(dataclass_instance)}


class _DataCollectionMeta(EnumMeta):
    """Metaclass that builds DataCollection class enums"""

    def __getattribute__(cls, item: str) -> Any:
        """This is executed whenever `DataCollection.SOMETHING` is called

        Extended method handles cases where a collection has been renamed. It provides a new collection and raises a
        deprecation warning.
        """
        if item in _RENAMED_COLLECTIONS:
            old_item = item
            item = _RENAMED_COLLECTIONS[old_item]

            message = (
                f"DataCollection.{old_item} had been renamed into DataCollection.{item}. Please switch to the "
                "new name as the old one will soon be removed."
            )
            warnings.warn(message, category=SHDeprecationWarning)

        return super().__getattribute__(item)

    def __call__(cls, value, *args, **kwargs):  # type: ignore
        """This is executed whenever `DataCollection('something')` is called

        This solves a problem of pickling a custom DataCollection and unpickling it in another process
        """
        if isinstance(value, DataCollectionDefinition) and value not in cls._value2member_map_ and value._name:
            cls._try_add_data_collection(value._name, value)

        return super().__call__(value, *args, **kwargs)


@dataclass(frozen=True)
class DataCollectionDefinition:
    """An immutable definition of a data collection

    Check `DataCollection.define` for more info about attributes of this class
    """

    # pylint: disable=too-many-instance-attributes
    api_id: Optional[str] = None
    catalog_id: Optional[str] = None
    wfs_id: Optional[str] = None
    service_url: Optional[str] = None
    collection_type: Optional[str] = None
    sensor_type: Optional[str] = None
    processing_level: Optional[str] = None
    swath_mode: Optional[str] = None
    polarization: Optional[str] = None
    resolution: Optional[str] = None
    orbit_direction: Optional[str] = None
    timeliness: Optional[str] = None
    bands: Optional[Tuple[Band, ...]] = None
    metabands: Optional[Tuple[Band, ...]] = None
    collection_id: Optional[str] = None
    is_timeless: bool = False
    has_cloud_coverage: bool = False
    dem_instance: Optional[str] = None
    # The following parameter is used to preserve custom DataCollection name during pickling and unpickling process:
    _name: Optional[str] = field(default=None, compare=False)

    def __post_init__(self):  # type: ignore
        """In case a list of bands or metabands has been given this makes sure to cast it into a tuple"""
        if isinstance(self.bands, list):
            object.__setattr__(self, "bands", tuple(self.bands))
        if isinstance(self.metabands, list):
            object.__setattr__(self, "metabands", tuple(self.metabands))

    def __repr__(self) -> str:
        """A nicer representation of parameters that define a data collection"""
        valid_params = {name: value for name, value in _shallow_asdict(self).items() if value is not None}
        params_repr = "\n  ".join(f"{name}: {value}" for name, value in valid_params.items() if name != "_name")
        return f"{self.__class__.__name__}(\n  {params_repr}\n)"

    def derive(self, **params: Any) -> "DataCollectionDefinition":
        """Create a new data collection definition from current definition and parameters that override current
        parameters

        :param params: Any of DataCollectionDefinition attributes
        :return: A new data collection definition
        """
        derived_params = _shallow_asdict(self)
        derived_params.update(params)

        return DataCollectionDefinition(**derived_params)


_RENAMED_COLLECTIONS = {  # DataCollection renaming for backwards-compatibility
    "LANDSAT15_L1": "LANDSAT_MSS_L1",
    "LANDSAT45_L1": "LANDSAT_TM_L1",
    "LANDSAT45_L2": "LANDSAT_TM_L2",
    "LANDSAT7_L1": "LANDSAT_ETM_L1",
    "LANDSAT7_L2": "LANDSAT_ETM_L2",
    "LANDSAT8": "LANDSAT_OT_L1",
    "LANDSAT8_L1": "LANDSAT_OT_L1",
    "LANDSAT8_L2": "LANDSAT_OT_L2",
}


class DataCollection(Enum, metaclass=_DataCollectionMeta):
    """An enum class for data collections

    It contains a number of predefined data collections, which are the most commonly used with Sentinel Hub service.
    Additionally, it also allows defining new data collections by specifying data collection parameters relevant for
    the service. Check `DataCollection.define` and similar methods for more.
    """

    SENTINEL2_L1C = DataCollectionDefinition(
        api_id="sentinel-2-l1c",
        catalog_id="sentinel-2-l1c",
        wfs_id="DSS1",
        service_url=ServiceUrl.MAIN,
        collection_type=_CollectionType.SENTINEL2,
        sensor_type=_SensorType.MSI,
        processing_level=_ProcessingLevel.L1C,
        bands=Bands.SENTINEL2_L1C,
        metabands=MetaBands.SENTINEL2_L1C,
        has_cloud_coverage=True,
    )
    SENTINEL2_L2A = DataCollectionDefinition(
        api_id="sentinel-2-l2a",
        catalog_id="sentinel-2-l2a",
        wfs_id="DSS2",
        service_url=ServiceUrl.MAIN,
        collection_type=_CollectionType.SENTINEL2,
        sensor_type=_SensorType.MSI,
        processing_level=_ProcessingLevel.L2A,
        bands=Bands.SENTINEL2_L2A,
        metabands=MetaBands.SENTINEL2_L2A,
        has_cloud_coverage=True,
    )

    SENTINEL1 = DataCollectionDefinition(
        api_id="sentinel-1-grd",
        catalog_id="sentinel-1-grd",
        wfs_id="DSS3",
        service_url=ServiceUrl.MAIN,
        collection_type=_CollectionType.SENTINEL1,
        sensor_type=_SensorType.C_SAR,
        processing_level=_ProcessingLevel.GRD,
        orbit_direction=OrbitDirection.BOTH,
        metabands=MetaBands.SENTINEL1,
    )
    SENTINEL1_IW = SENTINEL1.derive(
        swath_mode=_SwathMode.IW,
        polarization=_Polarization.DV,
        resolution=_Resolution.HIGH,
        bands=Bands.SENTINEL1_IW,
    )
    SENTINEL1_IW_ASC = SENTINEL1_IW.derive(
        orbit_direction=OrbitDirection.ASCENDING,
    )
    SENTINEL1_IW_DES = SENTINEL1_IW.derive(
        orbit_direction=OrbitDirection.DESCENDING,
    )
    SENTINEL1_EW = SENTINEL1.derive(
        swath_mode=_SwathMode.EW,
        polarization=_Polarization.DH,
        resolution=_Resolution.MEDIUM,
        bands=Bands.SENTINEL1_EW,
    )
    SENTINEL1_EW_ASC = SENTINEL1_EW.derive(
        orbit_direction=OrbitDirection.ASCENDING,
    )
    SENTINEL1_EW_DES = SENTINEL1_EW.derive(
        orbit_direction=OrbitDirection.DESCENDING,
    )
    SENTINEL1_EW_SH = SENTINEL1_EW.derive(
        polarization=_Polarization.SH,
        bands=Bands.SENTINEL1_EW_SH,
    )
    SENTINEL1_EW_SH_ASC = SENTINEL1_EW_SH.derive(
        orbit_direction=OrbitDirection.ASCENDING,
    )
    SENTINEL1_EW_SH_DES = SENTINEL1_EW_SH.derive(
        orbit_direction=OrbitDirection.DESCENDING,
    )

    DEM = DataCollectionDefinition(
        api_id="dem",
        service_url=ServiceUrl.MAIN,
        collection_type=_CollectionType.DEM,
        bands=Bands.DEM,
        metabands=MetaBands.DEM,
        is_timeless=True,
    )
    DEM_MAPZEN = DEM.derive(
        dem_instance="MAPZEN",
    )
    DEM_COPERNICUS_30 = DEM.derive(
        dem_instance="COPERNICUS_30",
    )
    DEM_COPERNICUS_90 = DEM.derive(
        dem_instance="COPERNICUS_90",
    )

    MODIS = DataCollectionDefinition(
        api_id="modis",
        catalog_id="modis",
        wfs_id="DSS5",
        service_url=ServiceUrl.USWEST,
        collection_type=_CollectionType.MODIS,
        bands=Bands.MODIS,
        metabands=MetaBands.MODIS,
    )

    LANDSAT_MSS_L1 = DataCollectionDefinition(
        api_id="landsat-mss-l1",
        catalog_id="landsat-mss-l1",
        wfs_id="DSS14",
        service_url=ServiceUrl.USWEST,
        collection_type=_CollectionType.LANDSAT_MSS,
        sensor_type=_SensorType.MSS,
        processing_level=_ProcessingLevel.L1,
        bands=Bands.LANDSAT_MSS_L1,
        metabands=MetaBands.LANDSAT_MSS_L1,
        has_cloud_coverage=True,
    )

    LANDSAT_TM_L1 = DataCollectionDefinition(
        api_id="landsat-tm-l1",
        catalog_id="landsat-tm-l1",
        wfs_id="DSS15",
        service_url=ServiceUrl.USWEST,
        collection_type=_CollectionType.LANDSAT_TM,
        sensor_type=_SensorType.TM,
        processing_level=_ProcessingLevel.L1,
        bands=Bands.LANDSAT_TM_L1,
        metabands=MetaBands.LANDSAT_TM_L1,
        has_cloud_coverage=True,
    )
    LANDSAT_TM_L2 = LANDSAT_TM_L1.derive(
        api_id="landsat-tm-l2",
        catalog_id="landsat-tm-l2",
        wfs_id="DSS16",
        processing_level=_ProcessingLevel.L2,
        bands=Bands.LANDSAT_TM_L2,
        metabands=MetaBands.LANDSAT_TM_L2,
    )

    LANDSAT_ETM_L1 = DataCollectionDefinition(
        api_id="landsat-etm-l1",
        catalog_id="landsat-etm-l1",
        wfs_id="DSS17",
        service_url=ServiceUrl.USWEST,
        collection_type=_CollectionType.LANDSAT_ETM,
        sensor_type=_SensorType.ETM,
        processing_level=_ProcessingLevel.L1,
        bands=Bands.LANDSAT_ETM_L1,
        metabands=MetaBands.LANDSAT_ETM_L1,
        has_cloud_coverage=True,
    )
    LANDSAT_ETM_L2 = LANDSAT_ETM_L1.derive(
        api_id="landsat-etm-l2",
        catalog_id="landsat-etm-l2",
        wfs_id="DSS18",
        processing_level=_ProcessingLevel.L2,
        bands=Bands.LANDSAT_ETM_L2,
        metabands=MetaBands.LANDSAT_ETM_L2,
    )

    LANDSAT_OT_L1 = DataCollectionDefinition(
        api_id="landsat-ot-l1",
        catalog_id="landsat-ot-l1",
        wfs_id="DSS12",
        service_url=ServiceUrl.USWEST,
        collection_type=_CollectionType.LANDSAT_OT,
        sensor_type=_SensorType.OLI_TIRS,
        processing_level=_ProcessingLevel.L1,
        bands=Bands.LANDSAT_OT_L1,
        metabands=MetaBands.LANDSAT_OT_L1,
        has_cloud_coverage=True,
    )
    LANDSAT_OT_L2 = LANDSAT_OT_L1.derive(
        api_id="landsat-ot-l2",
        catalog_id="landsat-ot-l2",
        wfs_id="DSS13",
        processing_level=_ProcessingLevel.L2,
        bands=Bands.LANDSAT_OT_L2,
        metabands=MetaBands.LANDSAT_OT_L2,
    )

    SENTINEL5P = DataCollectionDefinition(
        api_id="sentinel-5p-l2",
        catalog_id="sentinel-5p-l2",
        wfs_id="DSS7",
        service_url=ServiceUrl.CREODIAS,
        collection_type=_CollectionType.SENTINEL5P,
        sensor_type=_SensorType.TROPOMI,
        processing_level=_ProcessingLevel.L2,
        bands=Bands.SENTINEL5P,
        metabands=MetaBands.SENTINEL5P,
    )
    SENTINEL3_OLCI = DataCollectionDefinition(
        api_id="sentinel-3-olci",
        catalog_id="sentinel-3-olci",
        wfs_id="DSS8",
        service_url=ServiceUrl.CREODIAS,
        collection_type=_CollectionType.SENTINEL3,
        sensor_type=_SensorType.OLCI,
        processing_level=_ProcessingLevel.L1B,
        bands=Bands.SENTINEL3_OLCI,
        metabands=MetaBands.SENTINEL3_OLCI,
    )
    SENTINEL3_SLSTR = DataCollectionDefinition(
        api_id="sentinel-3-slstr",
        catalog_id="sentinel-3-slstr",
        wfs_id="DSS9",
        service_url=ServiceUrl.CREODIAS,
        collection_type=_CollectionType.SENTINEL3,
        sensor_type=_SensorType.SLSTR,
        processing_level=_ProcessingLevel.L1B,
        bands=Bands.SENTINEL3_SLSTR,
        metabands=MetaBands.SENTINEL3_SLSTR,
        has_cloud_coverage=True,
    )

    # pylint: disable=too-many-locals
    @classmethod
    def define(
        cls,
        name: str,
        *,
        api_id: Optional[str] = None,
        catalog_id: Optional[str] = None,
        wfs_id: Optional[str] = None,
        service_url: Optional[str] = None,
        collection_type: Optional[str] = None,
        sensor_type: Optional[str] = None,
        processing_level: Optional[str] = None,
        swath_mode: Optional[str] = None,
        polarization: Optional[str] = None,
        resolution: Optional[str] = None,
        orbit_direction: Optional[str] = None,
        timeliness: Optional[str] = None,
        bands: Optional[Tuple[Band, ...]] = None,
        metabands: Optional[Tuple[Band, ...]] = None,
        collection_id: Optional[str] = None,
        is_timeless: bool = False,
        has_cloud_coverage: bool = False,
        dem_instance: Optional[str] = None,
    ) -> "DataCollection":
        """Define a new data collection

        Note that all parameters, except `name` are optional. If a data collection definition won't be used for a
        certain use case (e.g. Process API, WFS, etc.), parameters for that use case don't have to be defined

        :param name: A name of a new data collection
        :param api_id: An ID to be used for Sentinel Hub Process API
        :param catalog_id: An ID to be used for Sentinel Hub Catalog API
        :param wfs_id: An ID to be used for Sentinel Hub WFS service
        :param service_url: A base URL of Sentinel Hub service deployment from where to download data. If it is not
            specified, a `sh_base_url` from a config will be used by default
        :param collection_type: A collection type
        :param sensor_type: A type of a satellite's sensor
        :param processing_level: A level of processing applied on satellite data
        :param swath_mode: A swath mode of SAR sensors
        :param polarization: A type of polarization
        :param resolution: A type of (Sentinel-1) resolution
        :param orbit_direction: A direction of satellite's orbit by which to filter satellite's data
        :param timeliness: A timeliness of data
        :param bands: Information about data collection bands
        :param metabands: Information about data collection metabands
        :param collection_id: An ID of a BYOC or BATCH collection
        :param is_timeless: `True` if a data collection can be filtered by time dimension and `False` otherwise
        :param has_cloud_coverage: `True` if data collection can be filtered by cloud coverage percentage and `False`
            otherwise
        :param dem_instance: one of the options listed in
            `DEM documentation <https://docs.sentinel-hub.com/api/latest/data/dem/#deminstance>`__
        :return: A new data collection
        """
        definition = DataCollectionDefinition(
            api_id=api_id,
            catalog_id=catalog_id,
            wfs_id=wfs_id,
            service_url=service_url,
            collection_type=collection_type,
            sensor_type=sensor_type,
            processing_level=processing_level,
            swath_mode=swath_mode,
            polarization=polarization,
            resolution=resolution,
            orbit_direction=orbit_direction,
            timeliness=timeliness,
            bands=bands,
            metabands=metabands,
            collection_id=collection_id,
            is_timeless=is_timeless,
            has_cloud_coverage=has_cloud_coverage,
            dem_instance=dem_instance,
            _name=name,
        )
        cls._try_add_data_collection(name, definition)
        return cls(definition)

    def define_from(self, name: str, **params: Any) -> "DataCollection":
        """Define a new data collection from an existing one

        :param name: A name of a new data collection
        :param params: Any parameter to override current data collection parameters
        :return: A new data collection
        """
        definition = self.value
        new_definition = definition.derive(**params, _name=name)

        self._try_add_data_collection(name, new_definition)
        return DataCollection(new_definition)

    @classmethod
    def _try_add_data_collection(cls, name: str, definition: DataCollectionDefinition) -> None:
        """Tries adding a new data collection definition. If the exact enum has already been defined then it won't do
        anything. However, if either a name or a definition has already been matched with another name or definition
        then it will raise an error.
        """
        is_name_defined = name in cls.__members__
        is_enum_defined = is_name_defined and cls.__members__[name].value == definition
        is_definition_defined = definition in cls._value2member_map_

        if is_enum_defined:
            return

        if not is_name_defined and not is_definition_defined:
            extend_enum(cls, name, definition)
            return

        if is_name_defined:
            raise ValueError(f"Data collection name '{name}' is already taken by another data collection")

        existing_collection = cls._value2member_map_[definition]
        raise ValueError(
            f"Data collection definition is already taken by {existing_collection}. Two different "
            "DataCollection enums cannot have the same definition."
        )

    @classmethod
    def define_byoc(cls, collection_id: str, **params: Any) -> "DataCollection":
        """Defines a BYOC data collection

        :param collection_id: An ID of a data collection
        :param params: Any parameter to override default BYOC data collection parameters
        :return: A new data collection
        """
        params["name"] = params.get("name", f"BYOC_{collection_id}")
        params["api_id"] = params.get("api_id", f"byoc-{collection_id}")
        params["catalog_id"] = params.get("catalog_id", f"byoc-{collection_id}")
        params["wfs_id"] = params.get("wfs_id", f"byoc-{collection_id}")
        params["collection_type"] = params.get("collection_type", _CollectionType.BYOC)
        params["collection_id"] = collection_id
        return cls.define(**params)

    @classmethod
    def define_batch(cls, collection_id: str, **params: Any) -> "DataCollection":
        """Defines a BATCH data collection

        :param collection_id: An ID of a data collection
        :param params: Any parameter to override default BATCH data collection parameters
        :return: A new data collection
        """
        params["name"] = params.get("name", f"BATCH_{collection_id}")
        params["api_id"] = params.get("api_id", f"batch-{collection_id}")
        params["catalog_id"] = params.get("catalog_id", f"batch-{collection_id}")
        params["wfs_id"] = params.get("wfs_id", f"batch-{collection_id}")
        params["collection_type"] = params.get("collection_type", _CollectionType.BATCH)
        params["collection_id"] = collection_id
        return cls.define(**params)

    @property
    def api_id(self) -> str:
        """Provides a Sentinel Hub Process API identifier or raises an error if it is not defined

        :return: An identifier
        :raises: ValueError
        """
        if self.value.api_id is None:
            raise ValueError(f"Data collection {self.name} is missing a Sentinel Hub Process API identifier")
        return self.value.api_id

    @property
    def catalog_id(self) -> str:
        """Provides a Sentinel Hub Catalog API identifier or raises an error if it is not defined

        :return: An identifier
        :raises: ValueError
        """
        if self.value.catalog_id is not None:
            return self.value.catalog_id
        if self.value.api_id is not None:
            # A fallback because Process API and Catalog API IDs should now be unified
            return self.value.api_id
        raise ValueError(f"Data collection {self.name} is missing a Sentinel Hub Catalog API identifier")

    @property
    def wfs_id(self) -> str:
        """Provides a Sentinel Hub WFS identifier or raises an error if it is not defined

        :return: An identifier
        :raises: ValueError
        """
        if self.value.wfs_id is None:
            raise ValueError(f"Data collection {self.name} is missing a Sentinel Hub WFS identifier")
        return self.value.wfs_id

    @property
    def bands(self) -> Tuple[Band, ...]:
        """Provides band information available for the data collection

        :return: A tuple of band info
        :raises: ValueError
        """
        if self.value.bands is None:
            raise ValueError(f"Data collection {self.name} does not define bands")
        return self.value.bands

    @property
    def metabands(self) -> Tuple[Band, ...]:
        """Provides metaband information available for the data collection

        :return: A tuple of metaband info
        :raises: ValueError
        """
        if self.value.metabands is None:
            raise ValueError(f"Data collection {self.name} does not define metabands")
        return self.value.metabands

    def __getattr__(self, item: str) -> Any:
        """The following insures that any attribute from DataCollectionDefinition, which is already not a
        property or an attribute of DataCollection, becomes an attribute of DataCollection
        """
        if not item.startswith("_") and hasattr(self, "value") and isinstance(self.value, DataCollectionDefinition):
            definition_dict = _shallow_asdict(self.value)
            if item in definition_dict:
                return definition_dict[item]

        return super().__getattribute__(item)

    @property
    def is_sentinel1(self) -> bool:
        """Checks if data collection is a Sentinel-1 collection type

        Example: ``DataCollection.SENTINEL1_IW.is_sentinel1``

        :return: `True` if collection is Sentinel-1 collection type and `False` otherwise
        """
        return self.collection_type == _CollectionType.SENTINEL1

    @property
    def is_byoc(self) -> bool:
        """Checks if data collection is a BYOC collection type

        :return: `True` if collection is a BYOC collection type and `False` otherwise
        """
        return self.collection_type == _CollectionType.BYOC

    @property
    def is_batch(self) -> bool:
        """Checks if data collection is a batch collection type

        :return: `True` if collection is a batch collection type and `False` otherwise
        """
        return self.collection_type == _CollectionType.BATCH

    def contains_orbit_direction(self, orbit_direction: str) -> bool:
        """Checks if a data collection contains given orbit direction

        :param orbit_direction: An orbit direction
        :return: `True` if data collection contains the orbit direction
        :return: bool
        """
        defined_direction = self.orbit_direction
        if defined_direction is None or defined_direction.upper() == OrbitDirection.BOTH:
            return True
        return orbit_direction.upper() == defined_direction.upper()

    @classmethod
    def get_available_collections(cls, config: Optional[SHConfig] = None) -> List["DataCollection"]:
        """Returns which data collections are available for configured Sentinel Hub OGC URL

        :param config: A custom instance of config class to override parameters from the saved configuration.
        :return: List of available data collections
        """
        if config is not None:
            warnings.warn(
                "Parameter config is deprecated, and will be removed in future.", category=SHDeprecationWarning
            )
        return list(cls)


DataSource = DataCollection


def handle_deprecated_data_source(
    data_collection: Optional[DataCollection],
    data_source: Optional[DataCollection],
    default: Optional[DataCollection] = None,
) -> DataCollection:
    """Joins parameters used to specify a data collection. In case data_source is given it raises a warning. In case
    both are given it raises an error. In case neither are given but there is a default collection it raises another
    warning.

    Note that this function is only temporary and will be removed in future package versions
    """
    if data_source is not None:
        warnings.warn("Parameter data_source is deprecated, use data_collection instead", category=SHDeprecationWarning)

    if data_collection is not None and data_source is not None:
        raise ValueError("Only one of the parameters data_collection and data_source should be given")

    if data_collection is None and data_source is None and default is not None:
        warnings.warn(
            "In the future please specify data_collection parameter, for now taking DataCollection.SENTINEL2_L1C",
            category=SHDeprecationWarning,
        )
        return default

    return data_collection or data_source  # type: ignore



"""
Module implementing DownloadRequest class
"""
import datetime as dt
import hashlib
import json
import os
import warnings
from typing import Any, Optional, Tuple

from ..constants import MimeType, RequestType
from ..exceptions import SHRuntimeWarning
from ..os_utils import sys_is_windows
from ..type_utils import JsonDict


class DownloadRequest:
    """A class defining a single download request.

    The class is a container with all parameters needed to execute a single download request and save or return
    downloaded data.
    """

    def __init__(
        self,
        *,
        url: Optional[str] = None,
        headers: Optional[JsonDict] = None,
        request_type: RequestType = RequestType.GET,
        post_values: Optional[JsonDict] = None,
        use_session: bool = False,
        data_type: MimeType = MimeType.RAW,
        save_response: bool = False,
        data_folder: Optional[str] = None,
        filename: Optional[str] = None,
        return_data: bool = True,
        **properties: Any,
    ):
        """
        :param url: A URL from where to download
        :param headers: Headers of an HTTP request
        :param request_type: Type of request, either GET or POST. Default is `RequestType.GET`
        :param post_values: A dictionary of values that will be sent with a POST request. Default is `None`
        :param use_session: A flag that specifies if the download request will require a session header
        :param data_type: An expected file format of downloaded data. Default is `MimeType.RAW`
        :param save_response: A flag defining if the downloaded data will be saved to disk. Default is `False`.
        :param data_folder: A folder path where the fetched data will be (or already is) saved. Default is `None`
        :param filename: A custom filename where the data will be saved. By default, data will be saved in a folder
            which name are hashed request parameters.
        :param return_data: A flag defining if the downloaded data will be returned as an output of download procedure.
            Default is `True`.
        :param properties: Any additional parameters.
        """
        self.url = url
        self.headers = headers or {}
        self.request_type = RequestType(request_type)
        self.post_values = post_values
        self.use_session = use_session

        self.data_type = MimeType(data_type)

        self.save_response = save_response
        self.data_folder = data_folder
        self.filename = filename
        self.return_data = return_data

        self.properties = properties

    def raise_if_invalid(self) -> None:
        """Method that raises an error if something is wrong with request parameters

        :raises: ValueError
        """
        if self.save_response and self.data_folder is None:
            raise ValueError(
                "Data folder is not specified. Please give a data folder name in the initialization of your request."
            )

    def get_request_params(self, include_metadata: bool = False) -> JsonDict:
        """Provides parameters that define the request in form of a dictionary

        :param include_metadata: A flag defining if also metadata parameters should be included, such as headers and
            current time
        :return: A dictionary of parameters
        """
        params = {"url": self.url, "payload": self.post_values}
        if include_metadata:
            params = {**params, "headers": self.headers, "timestamp": dt.datetime.now().isoformat()}
        return params

    def get_hashed_name(self) -> str:
        """It takes request url and payload and calculates a unique hashed string from them.

        :return: A hashed string
        """
        params = self.get_request_params(include_metadata=False)
        hashable = json.dumps(params)

        return hashlib.md5(hashable.encode("utf-8")).hexdigest()

    def get_relative_paths(self) -> Tuple[Optional[str], str]:
        """A method that calculates file paths relative to `data_folder`

        :return: Returns a pair of file paths, a request payload path and a response path. If request path is not
            defined it returns `None`.
        """

        if self.filename is not None:
            return None, self.filename

        hashed_name = self.get_hashed_name()

        request_path = os.path.join(hashed_name, "request.json")
        response_path = os.path.join(hashed_name, f"response.{self.data_type.extension}")

        return request_path, response_path

    def get_storage_paths(self) -> Tuple[Optional[str], Optional[str]]:
        """A method that calculates file paths where request payload and response will be saved.

        :return: Returns a pair of file paths, a request payload path and a response path. Each of them can also be
            `None` if it is not defined.
        """
        if self.data_folder is None:
            return None, None

        request_path, response_path = self.get_relative_paths()

        if request_path is not None:
            request_path = os.path.join(self.data_folder, request_path)
        response_path = os.path.join(self.data_folder, response_path)

        self._check_path(response_path)
        return request_path, response_path

    @staticmethod
    def _check_path(file_path: str) -> None:
        """Checks file path and warns about potential problems during saving"""
        message_problem = None
        if len(file_path) > 255 and sys_is_windows():
            message_problem = "File path"
        elif len(os.path.basename(file_path)) > 255:
            message_problem = "Filename of"

        if message_problem:
            message = (
                f"{message_problem} {file_path} is longer than 255 character which might cause an error while "
                "saving on disk"
            )
            warnings.warn(message, category=SHRuntimeWarning)



"""
Interface of
`Sentinel Hub Web Feature Service (WFS) <https://www.sentinel-hub.com/develop/api/ogc/standard-parameters/wfs/>`__.
"""
import datetime as dt
from typing import Iterable, List, Optional, Tuple, Union
from urllib.parse import urlencode

import shapely.geometry

from ..base import FeatureIterator
from ..config import SHConfig
from ..constants import CRS, MimeType, ServiceType, SHConstants
from ..data_collections import DataCollection
from ..download import SentinelHubDownloadClient
from ..geometry import BBox
from ..time_utils import parse_time, parse_time_interval, serialize_time
from ..type_utils import JsonDict, RawTimeIntervalType, RawTimeType


class WebFeatureService(FeatureIterator[JsonDict]):
    """Class for interaction with Sentinel Hub WFS service

    The class is an iterator over info about all available satellite tiles for requested parameters. It collects data
    from Sentinel Hub service only during the first iteration. During next iterations it returns already obtained data.
    The data is in the same order as returned by the service.

    For more info check `WFS documentation <https://www.sentinel-hub.com/develop/api/ogc/standard-parameters/wfs/>`__.
    """

    def __init__(
        self,
        bbox: BBox,
        time_interval: Union[RawTimeType, RawTimeIntervalType],
        *,
        data_collection: DataCollection,
        maxcc: float = 1.0,
        config: Optional[SHConfig] = None,
    ):
        """
        :param bbox: Bounding box of the requested image. Coordinates must be in the specified coordinate reference
            system.
        :param time_interval: interval with start and end date of the form YYYY-MM-DDThh:mm:ss or YYYY-MM-DD
        :param data_collection: A collection of requested satellite data
        :param maxcc: Maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is 1.0.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.config = config or SHConfig()
        self.config.raise_for_missing_instance_id()

        self.bbox = bbox

        self.latest_time_only = time_interval == SHConstants.LATEST
        if not self.latest_time_only:
            self.time_interval = parse_time_interval(time_interval)
        else:
            self.time_interval = dt.datetime(year=1985, month=1, day=1), dt.datetime.now()

        self.data_collection = data_collection
        self.maxcc = maxcc
        self.max_features_per_request = 1 if self.latest_time_only else self.config.max_wfs_records_per_query

        client = SentinelHubDownloadClient(config=self.config)
        url = self._build_service_url()
        params = self._build_request_params()

        super().__init__(client, url, params)
        self.next: int = 0

    def _build_service_url(self) -> str:
        """Creates a base URL for WFS service"""
        base_url = self.config.get_sh_ogc_url()

        if self.data_collection.service_url:
            base_url = base_url.replace(self.config.sh_base_url, self.data_collection.service_url)

        return f"{base_url}/{ServiceType.WFS.value}/{self.config.instance_id}"

    def _build_request_params(self) -> JsonDict:
        """Builds URL parameters for WFS service"""
        start_time, end_time = serialize_time(self.time_interval, use_tz=True)
        return {
            "SERVICE": ServiceType.WFS.value,
            "WARNINGS": False,
            "REQUEST": "GetFeature",
            "TYPENAMES": self.data_collection.wfs_id,
            "BBOX": str(self.bbox.reverse()) if self.bbox.crs is CRS.WGS84 else str(self.bbox),
            "OUTPUTFORMAT": MimeType.JSON.get_string(),
            "SRSNAME": self.bbox.crs.ogc_string(),
            "TIME": f"{start_time}/{end_time}",
            "MAXCC": 100.0 * self.maxcc,
            "MAXFEATURES": self.max_features_per_request,
        }

    def _fetch_features(self) -> Iterable[JsonDict]:
        """Collects data from WFS service"""
        params: JsonDict = {**self.params, "FEATURE_OFFSET": self.next}
        url = f"{self.url}?{urlencode(params)}"

        new_features = self.client.get_json_dict(url)["features"]

        if len(new_features) < self.max_features_per_request or self.latest_time_only:
            self.finished = True
        else:
            self.next += self.max_features_per_request

        is_sentinel1 = self.data_collection.is_sentinel1
        new_features = [
            feature_info
            for feature_info in new_features
            if not is_sentinel1 or self._sentinel1_product_check(feature_info)
        ]
        return new_features

    def get_dates(self) -> List[Optional[dt.date]]:
        """Returns a list of acquisition times from tile info data

        :return: List of acquisition times in the order returned by WFS service.
        """
        tile_dates: List[Optional[dt.date]] = []

        for tile_info in self:
            if not tile_info["properties"]["date"]:  # could be True for custom (BYOC) data collections
                tile_dates.append(None)
            else:
                date_str = tile_info["properties"]["date"]
                time_str = tile_info["properties"]["time"]
                tile_dates.append(parse_time(f"{date_str}T{time_str}"))

        return tile_dates

    def get_geometries(self) -> List[shapely.geometry.MultiPolygon]:
        """Returns a list of geometries from tile info data

        :return: List of multipolygon geometries in the order returned by WFS service.
        """
        return [shapely.geometry.shape(tile_info["geometry"]) for tile_info in self]

    def get_tiles(self) -> List[Tuple[str, str, int]]:
        """Returns list of tiles with tile name, date and AWS index

        :return: List of tiles in form of (tile_name, date, aws_index)
        """
        return [self._parse_tile_url(tile_info["properties"]["path"]) for tile_info in self]

    @staticmethod
    def _parse_tile_url(tile_url: str) -> Tuple[str, str, int]:
        """Extracts tile name, data and AWS index from tile URL

        :param tile_url: Location of tile at AWS
        :return: Tuple in a form (tile_name, date, aws_index)
        """
        props = tile_url.rsplit("/", 7)
        return "".join(props[1:4]), "-".join(props[4:7]), int(props[7])

    def _sentinel1_product_check(self, tile_info: JsonDict) -> bool:
        """Checks if Sentinel-1 tile info match the data collection definition"""
        product_id = tile_info["properties"]["id"]
        props = product_id.split("_")
        swath_mode, resolution, polarization = props[1], props[2][3], props[3][2:4]
        orbit_direction = tile_info["properties"].get("orbitDirection", "")

        if not (swath_mode in ["IW", "EW"] and resolution in ["M", "H"] and polarization in ["DV", "DH", "SV", "SH"]):
            raise ValueError(f"Unknown Sentinel-1 tile type: {product_id}")

        return (
            (swath_mode == self.data_collection.swath_mode or self.data_collection.swath_mode is None)
            and (polarization == self.data_collection.polarization or self.data_collection.polarization is None)
            and (resolution == self.data_collection.resolution[0] or self.data_collection.resolution is None)
            and self.data_collection.contains_orbit_direction(orbit_direction)
        )



"""
Module for working with large geographical areas
"""
import itertools
import json
import math
import os
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Iterable, List, Optional, Tuple, TypeVar, Union, cast

import shapely
import shapely.geometry
import shapely.ops
from shapely.geometry import GeometryCollection, MultiPolygon, Polygon
from shapely.geometry.base import BaseGeometry

from .api import BatchRequest, SentinelHubBatch, SentinelHubCatalog
from .config import SHConfig
from .constants import CRS
from .data_collections import DataCollection
from .geo_utils import transform_point
from .geometry import BBox, BBoxCollection, Geometry, _BaseGeometry
from .type_utils import JsonDict

T = TypeVar("T", float, int)


class AreaSplitter(metaclass=ABCMeta):
    """Abstract class for splitter classes. It implements common methods used for splitting large area into smaller
    parts.
    """

    def __init__(
        self,
        shape_list: Iterable[Union[Polygon, MultiPolygon, _BaseGeometry]],
        crs: CRS,
        reduce_bbox_sizes: bool = False,
    ):
        """
        :param shape_list: A list of geometrical shapes describing the area of interest
        :param crs: Coordinate reference system of the shapes in `shape_list`
        :param reduce_bbox_sizes: If `True` it will reduce the sizes of bounding boxes so that they will tightly fit
            the given geometry in `shape_list`.
        """
        self.crs = CRS(crs)
        self.shape_list = [self._parse_shape(shape, crs) for shape in shape_list]
        self.area_shape = self._join_shape_list(self.shape_list)
        self.reduce_bbox_sizes = reduce_bbox_sizes

        self.area_bbox = self.get_area_bbox()
        self.bbox_list, self.info_list = self._make_split()

    @staticmethod
    def _parse_shape(shape: Union[Polygon, MultiPolygon, _BaseGeometry], crs: CRS) -> Union[Polygon, MultiPolygon]:
        """Helper method for parsing input shapes"""
        if isinstance(shape, (Polygon, MultiPolygon)):
            return shape
        if isinstance(shape, _BaseGeometry):
            return shape.transform(crs).geometry
        raise ValueError(
            f"The list of shapes must contain shapes of types {Polygon}, {MultiPolygon} or subtype of {_BaseGeometry}"
        )

    @staticmethod
    def _join_shape_list(shape_list: List[Union[Polygon, MultiPolygon]]) -> MultiPolygon:
        """Joins a list of shapes together into one shape

        :param shape_list: A list of geometrical shapes describing the area of interest
        :return: A multipolygon which is a union of shapes in given list
        """
        if shapely.__version__ >= "1.8.0":
            return shapely.ops.unary_union(shape_list)
        return shapely.ops.cascaded_union(shape_list)

    @abstractmethod
    def _make_split(self) -> Tuple[List[BBox], List[Dict[str, object]]]:
        """The abstract method where the splitting will happen"""

    def get_bbox_list(
        self,
        crs: Optional[CRS] = None,
        buffer: Union[None, float, Tuple[float, float]] = None,
        reduce_bbox_sizes: Optional[bool] = None,
    ) -> List[BBox]:
        """Returns a list of bounding boxes that are the result of the split

        :param crs: Coordinate reference system in which the bounding boxes should be returned. If `None` the CRS will
            be the default CRS of the splitter.
        :param buffer: A percentage of each BBox size increase. This will cause neighbouring bounding boxes to overlap.
        :param reduce_bbox_sizes: If `True` it will reduce the sizes of bounding boxes so that they will tightly
            fit the given geometry in `shape_list`. This overrides the same parameter from constructor
        :return: List of bounding boxes
        """
        bbox_list = self.bbox_list
        if buffer:
            bbox_list = [bbox.buffer(buffer, relative=True) for bbox in bbox_list]

        if reduce_bbox_sizes is None:
            reduce_bbox_sizes = self.reduce_bbox_sizes
        if reduce_bbox_sizes:
            bbox_list = self._reduce_sizes(bbox_list)

        if crs:
            return [bbox.transform(crs) for bbox in bbox_list]
        return bbox_list

    def get_geometry_list(self) -> List[Union[Polygon, MultiPolygon]]:
        """For each bounding box an intersection with the shape of entire given area is calculated. CRS of the returned
        shapes is the same as CRS of the given area.

        :return: List of polygons or multipolygons corresponding to the order of bounding boxes
        """
        return [self._intersection_area(bbox) for bbox in self.bbox_list]

    def get_info_list(self) -> List[Dict[str, object]]:
        """Returns a list of dictionaries containing information about bounding boxes obtained in split. The order in
        the list matches the order of the list of bounding boxes.

        :return: List of dictionaries
        """
        return self.info_list

    def get_area_shape(self) -> MultiPolygon:
        """Returns a single shape of entire area described with `shape_list` parameter

        :return: A multipolygon which is a union of shapes describing the area
        """
        return self.area_shape

    def get_area_bbox(self, crs: Optional[CRS] = None) -> BBox:
        """Returns a bounding box of the entire area

        :param crs: Coordinate reference system in which the bounding box should be returned. If `None` the CRS will
            be the default CRS of the splitter.
        :return: A bounding box of the area defined by the `shape_list`
        """
        bbox_list = [BBox(shape.bounds, crs=self.crs) for shape in self.shape_list]
        area_min_x = min(bbox.lower_left[0] for bbox in bbox_list)
        area_min_y = min(bbox.lower_left[1] for bbox in bbox_list)
        area_max_x = max(bbox.upper_right[0] for bbox in bbox_list)
        area_max_y = max(bbox.upper_right[1] for bbox in bbox_list)
        bbox = BBox([area_min_x, area_min_y, area_max_x, area_max_y], crs=self.crs)
        if crs is None:
            return bbox
        return bbox.transform(crs)

    def _intersects_area(self, bbox: BBox) -> bool:
        """Checks if the bounding box intersects the entire area

        :param bbox: A bounding box
        :return: `True` if bbox intersects the entire area else False
        """
        return self._bbox_to_area_polygon(bbox).intersects(self.area_shape)

    def _intersection_area(self, bbox: BBox) -> Union[Polygon, MultiPolygon]:
        """Calculates the intersection of a given bounding box and the entire area

        :param bbox: A bounding box
        :return: A shape of intersection
        """
        return self._bbox_to_area_polygon(bbox).intersection(self.area_shape)

    def _bbox_to_area_polygon(self, bbox: BBox) -> Polygon:
        """Transforms bounding box into a polygon object in the area CRS.

        :param bbox: A bounding box
        :return: A polygon
        """
        projected_bbox = bbox.transform(self.crs)
        return projected_bbox.geometry

    def _reduce_sizes(self, bbox_list: List[BBox]) -> List[BBox]:
        """Reduces sizes of bounding boxes"""
        return [BBox(self._intersection_area(bbox).bounds, self.crs).transform(bbox.crs) for bbox in bbox_list]


class BBoxSplitter(AreaSplitter):
    """A tool that splits the given area into smaller parts. Given the area it calculates its bounding box and splits
    it into smaller bounding boxes of equal size. Then it filters out the bounding boxes that do not intersect the
    area. If specified by user it can also reduce the sizes of the remaining bounding boxes to best fit the area.
    """

    def __init__(
        self,
        shape_list: Iterable[Union[Polygon, MultiPolygon, _BaseGeometry]],
        crs: CRS,
        split_shape: Union[int, Tuple[int, int]],
        **kwargs: Any,
    ):
        """
        :param shape_list: A list of geometrical shapes describing the area of interest
        :param crs: Coordinate reference system of the shapes in `shape_list`
        :param split_shape: Parameter that describes the shape in which the area bounding box will be split.
            It can be a tuple of the form `(n, m)` which means the area bounding box will be split into `n` columns
            and `m` rows. It can also be a single integer `n` which is the same as `(n, n)`.
        :param reduce_bbox_sizes: If `True` it will reduce the sizes of bounding boxes so that they will tightly fit
            the given area geometry from `shape_list`.
        """
        self.split_shape = _parse_to_pair(split_shape, allowed_types=(int,), param_name="split_shape")
        super().__init__(shape_list, crs, **kwargs)

    def _make_split(self) -> Tuple[List[BBox], List[Dict[str, object]]]:
        columns, rows = self.split_shape
        bbox_partition = self.area_bbox.get_partition(num_x=columns, num_y=rows)

        bbox_list, info_list = [], []
        for i, j in itertools.product(range(columns), range(rows)):
            if self._intersects_area(bbox_partition[i][j]):
                bbox_list.append(bbox_partition[i][j])

                info = {"parent_bbox": self.area_bbox, "index_x": i, "index_y": j}
                info_list.append(info)

        return bbox_list, info_list


class OsmSplitter(AreaSplitter):
    """A tool that splits the given area into smaller parts. For the splitting it uses Open Street Map (OSM) grid on
    the specified zoom level. It calculates bounding boxes of all OSM tiles that intersect the area. If specified by
    user it can also reduce the sizes of the remaining bounding boxes to best fit the area.
    """

    def __init__(
        self,
        shape_list: Iterable[Union[Polygon, MultiPolygon, _BaseGeometry]],
        crs: CRS,
        zoom_level: int,
        **kwargs: Any,
    ):
        """
        :param shape_list: A list of geometrical shapes describing the area of interest
        :param crs: Coordinate reference system of the shapes in `shape_list`
        :param zoom_level: A zoom level defined by OSM. Level 0 is entire world, level 1 splits the world into
            4 parts, etc.
        :param reduce_bbox_sizes: If `True` it will reduce the sizes of bounding boxes so that they will tightly fit
            the given area geometry from `shape_list`.
        """
        self._POP_WEB_MAX = transform_point((180, 0), CRS.WGS84, CRS.POP_WEB)[0]  # pylint: disable=invalid-name

        self.zoom_level = zoom_level
        super().__init__(shape_list, crs, **kwargs)

    def _make_split(self) -> Tuple[List[BBox], List[Dict[str, object]]]:
        self.area_bbox = self.get_area_bbox(CRS.POP_WEB)
        self._check_area_bbox()

        bbox_list, info_list = self._recursive_split(self.get_world_bbox(), 0, 0, 0)

        for i, bbox in enumerate(bbox_list):
            bbox_list[i] = bbox.transform(self.crs)
        return bbox_list, info_list

    def _check_area_bbox(self) -> None:
        """The method checks if the area bounding box is completely inside the OSM grid. That means that its latitudes
        must be contained in the interval (-85.0511, 85.0511)

        :raises: ValueError
        """
        for coord in self.area_bbox:
            if abs(coord) > self._POP_WEB_MAX:
                raise ValueError(
                    "OsmTileSplitter only works for areas which have latitude in interval (-85.0511, 85.0511)"
                )

    def get_world_bbox(self) -> BBox:
        """Creates a bounding box of the entire world in EPSG: 3857

        :return: Bounding box of entire world
        """
        return BBox((-self._POP_WEB_MAX, -self._POP_WEB_MAX, self._POP_WEB_MAX, self._POP_WEB_MAX), crs=CRS.POP_WEB)

    def _recursive_split(
        self,
        bbox: BBox,
        zoom_level: int,
        column: int,
        row: int,
    ) -> Tuple[List[BBox], List[Dict[str, object]]]:
        """Method that recursively creates bounding boxes of OSM grid that intersect the area.

        :param bbox: Bounding box
        :param zoom_level: OSM zoom level
        :param column: Column in the OSM grid
        :param row: Row in the OSM grid
        """
        if zoom_level == self.zoom_level:
            return [bbox], [{"zoom_level": zoom_level, "index_x": column, "index_y": row}]

        bbox_list, info_list = [], []
        bbox_partition = bbox.get_partition(num_x=2, num_y=2)
        for i, j in itertools.product(range(2), range(2)):
            if self._intersects_area(bbox_partition[i][j]):
                bboxes, infos = self._recursive_split(
                    bbox_partition[i][j], zoom_level + 1, 2 * column + i, 2 * row + 1 - j
                )
                bbox_list.extend(bboxes)
                info_list.extend(infos)

        return bbox_list, info_list


class TileSplitter(AreaSplitter):
    """A splitter that uses Sentinel Hub Catalog API to obtain geometries of the original tiling grid of a given
    data collection. Additionally, it can further split these geometries into smaller parts.
    """

    _CATALOG_FILTER = {
        "include": ["id", "geometry", "properties.datetime", "properties.proj:bbox", "properties.proj:epsg"],
        "exclude": [],
    }

    def __init__(
        self,
        shape_list: Iterable[Union[Polygon, MultiPolygon, _BaseGeometry]],
        crs: CRS,
        time_interval: Tuple[str, str],
        data_collection: DataCollection,
        tile_split_shape: Union[int, Tuple[int, int]] = 1,
        config: Optional[SHConfig] = None,
        **kwargs: Any,
    ):
        """
        :param shape_list: A list of geometrical shapes describing the area of interest
        :param crs: Coordinate reference system of the shapes in `shape_list`
        :param time_interval: Interval with start and end date of the form YYYY-MM-DDThh:mm:ss or YYYY-MM-DD
        :param data_collection: A satellite data collection
        :param tile_split_shape: Parameter that describes the shape in which the satellite tile bounding boxes will be
            split. It can be a tuple of the form `(n, m)` which means the tile bounding boxes will be
            split into `n` columns and `m` rows. It can also be a single integer `n` which is the same
            as `(n, n)`.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        :param kwargs: Parameters that are propagated to the base `AreaSplitter` class
        """
        self.time_interval = time_interval
        self.tile_split_shape = tile_split_shape
        self.data_collection = data_collection

        sh_config = config or SHConfig()
        if data_collection.service_url:
            sh_config = sh_config.copy()
            sh_config.sh_base_url = data_collection.service_url
        self.catalog = SentinelHubCatalog(config=sh_config)
        super().__init__(shape_list, crs, **kwargs)

    def _make_split(self) -> Tuple[List[BBox], List[Dict[str, object]]]:
        tile_dict: Dict[Tuple[Tuple[float, ...], int], Dict[str, Any]] = {}

        search_iterator = self.catalog.search(
            self.data_collection, time=self.time_interval, bbox=self.area_bbox, fields=self._CATALOG_FILTER
        )

        timestamps = search_iterator.get_timestamps()
        geometry_list = search_iterator.get_geometries()
        for tile_info, (timestamp, geometry) in zip(search_iterator, zip(timestamps, geometry_list)):
            tile_properties = tile_info["properties"]
            bbox = BBox(tile_properties["proj:bbox"], crs=tile_properties["proj:epsg"])
            bbox_hash = tuple(bbox), bbox.crs.epsg

            if bbox_hash not in tile_dict:
                tile_dict[bbox_hash] = {"bbox": bbox, "timestamps": [], "ids": [], "geometries": []}
            tile_dict[bbox_hash]["timestamps"].append(timestamp)
            tile_dict[bbox_hash]["ids"].append(tile_info["id"])
            tile_dict[bbox_hash]["geometries"].append(geometry)

        bbox_list, info_list = [], []

        for tile_info in tile_dict.values():
            if not self._intersects_area(tile_info["bbox"]):
                continue

            tile_bbox: BBox = tile_info["bbox"]
            bbox_splitter = BBoxSplitter([tile_bbox.geometry], tile_bbox.crs, split_shape=self.tile_split_shape)

            for bbox, info in zip(bbox_splitter.get_bbox_list(), bbox_splitter.get_info_list()):
                if self._intersects_area(bbox):
                    bbox_list.append(bbox)

                    info["ids"] = tile_info["ids"]
                    info["timestamps"] = tile_info["timestamps"]
                    info_list.append(info)

        return bbox_list, info_list


class CustomGridSplitter(AreaSplitter):
    """Splitting class which can split according to given custom collection of bounding boxes"""

    def __init__(
        self,
        shape_list: Iterable[Union[Polygon, MultiPolygon, _BaseGeometry]],
        crs: CRS,
        bbox_grid: Union[List[BBox], BBoxCollection],
        bbox_split_shape: Union[int, Tuple[int, int]] = 1,
        **kwargs: Any,
    ):
        """
        :param shape_list: A list of geometrical shapes describing the area of interest
        :param crs: Coordinate reference system of the shapes in `shape_list`
        :param bbox_grid: A collection of bounding boxes defining a grid of splitting. All of them have to be in the
            same CRS.
        :param bbox_split_shape: Parameter that describes the shape in which each of the bounding boxes in the given
            grid will be split. It can be a tuple of the form `(n, m)` which means the tile bounding boxes will be
            split into `n` columns and `m` rows. It can also be a single integer `n` which is the same as `(n, n)`.
        :param reduce_bbox_sizes: If `True` it will reduce the sizes of bounding boxes so that they will tightly fit
            the given geometry in `shape_list`.
        """
        self.bbox_grid = self._parse_bbox_grid(bbox_grid)
        self.bbox_split_shape = bbox_split_shape
        super().__init__(shape_list, crs, **kwargs)

    @staticmethod
    def _parse_bbox_grid(bbox_grid: Union[List[BBox], BBoxCollection]) -> BBoxCollection:
        """Helper method for parsing bounding box grid. It will try to parse it into `BBoxCollection`"""
        if isinstance(bbox_grid, BBoxCollection):
            return bbox_grid

        if isinstance(bbox_grid, list):
            return BBoxCollection(bbox_grid)

        raise ValueError(f"Parameter 'bbox_grid' should be an instance of {BBoxCollection}")

    def _make_split(self) -> Tuple[List[BBox], List[Dict[str, object]]]:
        bbox_list: List[BBox] = []
        info_list: List[Dict[str, object]] = []

        for grid_idx, grid_bbox in enumerate(self.bbox_grid):
            if self._intersects_area(grid_bbox):
                bbox_splitter = BBoxSplitter([grid_bbox.geometry], grid_bbox.crs, split_shape=self.bbox_split_shape)

                for bbox, info in zip(bbox_splitter.get_bbox_list(), bbox_splitter.get_info_list()):
                    if self._intersects_area(bbox):
                        info["grid_index"] = grid_idx

                        bbox_list.append(bbox)
                        info_list.append(info)

        return bbox_list, info_list


class BaseUtmSplitter(AreaSplitter):
    """Base splitter that returns bboxes of fixed size aligned to UTM zones or UTM grid tiles as defined by the MGRS

    The generated bounding box grid will have coordinates in form of
    `(N * bbox_size_x + offset_x, M * bbox_size_y + offset_y)`
    """

    def __init__(
        self,
        shape_list: Iterable[Union[Polygon, MultiPolygon, _BaseGeometry]],
        crs: CRS,
        bbox_size: Union[float, Tuple[float, float]],
        offset: Union[None, Tuple[float, float]] = None,
    ):
        """
        :param shape_list: A list of geometrical shapes describing the area of interest
        :param crs: Coordinate reference system of the shapes in `shape_list`
        :param bbox_size: A size of generated bounding boxes in horizontal and vertical directions in meters. If a
            single value is given that will be interpreted as (value, value).
        :param offset: Bounding box offset in horizontal and vertical directions in meters.
        """
        self.bbox_size = _parse_to_pair(bbox_size, allowed_types=(int, float), param_name="bbox_size")

        self.offset = _parse_to_pair(offset or 0.0, allowed_types=(int, float), param_name="offset")

        self.utm_grid = self._get_utm_polygons()
        super().__init__(shape_list, crs)

    @abstractmethod
    def _get_utm_polygons(self) -> List[Tuple[BaseGeometry, Dict[str, Any]]]:
        raise NotImplementedError

    @staticmethod
    def _get_utm_from_props(utm_dict: Dict[str, Any]) -> CRS:
        """Return the UTM CRS corresponding to the UTM described by the properties dictionary

        :param utm_dict: Dictionary reporting name of the UTM zone and MGRS grid
        :return: UTM coordinate reference system
        """
        hemisphere_digit = 6 if utm_dict["direction"] == "N" else 7
        zone_number = utm_dict["zone"]
        return CRS(f"32{hemisphere_digit}{zone_number:02d}")

    def _align_bbox_to_size(self, bbox: BBox) -> BBox:
        """Align input bbox coordinates to be multiples of the bbox size

        :param bbox: Bounding box in UTM coordinates
        :return: BBox objects with coordinates multiples of the bbox size
        """
        size_x, size_y = self.bbox_size
        offset_x, offset_y = self.offset
        lower_left_x, lower_left_y = bbox.lower_left

        aligned_x = math.floor((lower_left_x - offset_x) / size_x) * size_x + offset_x
        aligned_y = math.floor((lower_left_y - offset_y) / size_y) * size_y + offset_y

        return BBox(((aligned_x, aligned_y), bbox.upper_right), crs=bbox.crs)

    def _make_split(self) -> Tuple[List[BBox], List[Dict[str, object]]]:
        """Split each UTM grid into equally sized bboxes in correct UTM zone"""
        size_x, size_y = self.bbox_size
        bbox_list: List[BBox] = []
        info_list: List[Dict[str, object]] = []

        index = 0
        shape_geometry = Geometry(self.area_shape, self.crs).transform(CRS.WGS84)

        for utm_cell in self.utm_grid:
            utm_cell_geom, utm_cell_prop = utm_cell
            # the UTM MGRS grid definition contains four 0 zones at the poles (0A, 0B, 0Y, 0Z)
            if utm_cell_prop["zone"] == 0:
                continue
            utm_crs = self._get_utm_from_props(utm_cell_prop)
            cell_info = dict(
                crs=utm_crs.name,
                utm_zone=str(utm_cell_prop["zone"]).zfill(2),
                utm_row=utm_cell_prop["row"],
                direction=utm_cell_prop["direction"],
            )

            intersection = utm_cell_geom.intersection(shape_geometry.geometry)

            if not intersection.is_empty and isinstance(intersection, GeometryCollection):
                intersection = MultiPolygon(
                    geo_object for geo_object in intersection if isinstance(geo_object, (Polygon, MultiPolygon))
                )

            if not intersection.is_empty:
                intersection = Geometry(intersection, CRS.WGS84).transform(utm_crs)

                bbox_partition = self._align_bbox_to_size(intersection.bbox).get_partition(size_x=size_x, size_y=size_y)

                for i, column in enumerate(bbox_partition):
                    for j, part in enumerate(column):
                        if part.geometry.intersects(intersection.geometry):
                            bbox_list.append(part)
                            info_list.append(dict(**cell_info, index=index, index_x=i, index_y=j))
                            index += 1

        return bbox_list, info_list

    def get_bbox_list(self, buffer: Union[None, float, Tuple[float, float]] = None) -> List[BBox]:  # type: ignore
        """Get list of bounding boxes.

        The CRS is fixed to the computed UTM CRS. This BBox splitter does not support reducing size of output
        bounding boxes

        :param buffer: A percentage of each BBox size increase. This will cause neighbouring bounding boxes to overlap.
        :return: List of bounding boxes
        """
        return super().get_bbox_list(buffer=buffer)


class UtmGridSplitter(BaseUtmSplitter):
    """Splitter that returns bounding boxes of fixed size aligned to the UTM MGRS grid"""

    def _get_utm_polygons(self) -> List[Tuple[BaseGeometry, Dict[str, Any]]]:
        """Find UTM grid zones overlapping with input area shape

        :return: List of geometries and properties of UTM grid zones overlapping with input area shape
        """
        # file downloaded from faculty.baruch.cuny.edu/geoportal/data/esri/world/utmzone.zip
        utm_grid_filename = os.path.join(os.path.dirname(__file__), ".utmzones.geojson")

        if not os.path.isfile(utm_grid_filename):
            raise IOError(f"UTM grid definition file does not exist: {os.path.abspath(utm_grid_filename)}")

        with open(utm_grid_filename) as utm_grid_file:
            utm_grid = json.load(utm_grid_file)["features"]

        utm_geom_list = [shapely.geometry.shape(utm_zone["geometry"]) for utm_zone in utm_grid]
        utm_prop_list = [
            dict(
                zone=utm_zone["properties"]["ZONE"],
                row=utm_zone["properties"]["ROW_"],
                direction="N" if utm_zone["properties"]["ROW_"] >= "N" else "S",
            )
            for utm_zone in utm_grid
        ]

        return list(zip(utm_geom_list, utm_prop_list))


class UtmZoneSplitter(BaseUtmSplitter):
    """Splitter that returns bounding boxes of fixed size aligned to the equator and the UTM zones."""

    LNG_MIN, LNG_MAX, LNG_UTM = -180, 180, 6
    LAT_MIN, LAT_MAX, LAT_EQ = -80, 84, 0

    def _get_utm_polygons(self) -> List[Tuple[BaseGeometry, Dict[str, Any]]]:
        """Find UTM zones overlapping with input area shape

        The returned geometry corresponds to a triangle ranging from the equator to the North/South Pole

        :return: List of geometries and properties of UTM zones overlapping with input area shape
        """
        utm_geom_list = []
        for lat in [(self.LAT_EQ, self.LAT_MAX), (self.LAT_MIN, self.LAT_EQ)]:
            for lng in range(self.LNG_MIN, self.LNG_MAX, self.LNG_UTM):
                points = []
                # A new point is added per each degree - this is inline with geometries used by UtmGridSplitter
                # In the future the number of points will be calculated according to bbox_size parameter
                for degree in range(lat[0], lat[1]):
                    points.append((lng, degree))
                for degree in range(lng, lng + self.LNG_UTM):
                    points.append((degree, lat[1]))
                for degree in range(lat[1], lat[0], -1):
                    points.append((lng + self.LNG_UTM, degree))
                for degree in range(lng + self.LNG_UTM, lng, -1):
                    points.append((degree, lat[0]))

                utm_geom_list.append(Polygon(points))

        utm_prop_list = [
            dict(zone=zone, row="", direction=direction) for direction in ["N", "S"] for zone in range(1, 61)
        ]

        return list(zip(utm_geom_list, utm_prop_list))


class BatchSplitter(AreaSplitter):
    """A splitter that obtains split bounding boxes from Sentinel Hub Batch API"""

    def __init__(
        self,
        *,
        request_id: Optional[str] = None,
        batch_request: Optional[BatchRequest] = None,
        config: Optional[SHConfig] = None,
    ):
        """
        :param request_id: An ID of a batch request
        :param batch_request: A batch request object. It is an alternative to the `request_id` parameter
        :param config: A configuration object with credentials and information about which service deployment to
            use.
        """
        self.batch_client = SentinelHubBatch(config=config)

        if batch_request is None:
            if request_id is None:
                raise ValueError("One of the parameters request_id and batch_request has to be given")
            batch_request = self.batch_client.get_request(request_id)

        self.batch_request = batch_request
        self.tile_size = self._get_tile_size()
        self.tile_buffer = self._get_tile_buffer()

        batch_geometry = batch_request.geometry
        super().__init__([batch_geometry.geometry], batch_geometry.crs)

    def _get_tile_size(self) -> Tuple[float, float]:
        """Collects a tile size from the tiling grid info in units of the grid CRS."""
        tiling_grid_id = self.batch_request.tiling_grid["id"]
        grid_info = self.batch_client.get_tiling_grid(tiling_grid_id)

        return grid_info["properties"]["tileWidth"], grid_info["properties"]["tileHeight"]

    def _get_tile_buffer(self) -> Tuple[float, float]:
        """Calculates tile buffer in units of the grid CRS."""
        grid_info = self.batch_request.tiling_grid
        resolution = grid_info["resolution"]
        return grid_info.get("bufferX", 0) * resolution, grid_info.get("bufferY", 0) * resolution

    def _make_split(self) -> Tuple[List[BBox], List[Dict[str, object]]]:
        """This method actually loads bounding boxes from the service and prepares the lists"""
        tile_info_list = list(self.batch_client.iter_tiles(self.batch_request))

        bbox_list = [self._reconstruct_bbox(tile_info) for tile_info in tile_info_list]
        info_list = [
            {key: value for key, value in tile_info.items() if key != "geometry"} for tile_info in tile_info_list
        ]

        return bbox_list, info_list

    def _reconstruct_bbox(self, tile_info: JsonDict) -> BBox:
        """Reconstructs a bounding box from tile and grid properties."""
        tile_crs = CRS(tile_info["origin"]["crs"]["properties"]["name"])

        upper_left_corner = tile_info["origin"]["coordinates"]
        width, height = self.tile_size

        return BBox(
            [
                upper_left_corner[0] - self.tile_buffer[0],
                upper_left_corner[1] - height - self.tile_buffer[1],
                upper_left_corner[0] + width + self.tile_buffer[0],
                upper_left_corner[1] + self.tile_buffer[1],
            ],
            tile_crs,
        )


def _parse_to_pair(
    parameter: Union[T, Tuple[T, T]], allowed_types: Tuple[type, ...], param_name: str = ""
) -> Tuple[T, T]:
    """Parses the parameters defining the splitting of the BBox."""

    if isinstance(parameter, (tuple, list)) and len(parameter) == 2:
        split_x, split_y = parameter
        if isinstance(split_x, allowed_types) and isinstance(split_y, allowed_types):
            return split_x, split_y

    parameter = cast(T, parameter)
    if isinstance(parameter, allowed_types):
        return parameter, parameter

    raise ValueError(
        f"Parameter {param_name} must be a single instance or a pair, with allowed types {allowed_types}, but"
        f" {parameter} was given"
    )
